"""The components of the derivative of the entropy for the Choo and Siow homoskedastic model.
"""

from typing import cast

import numpy as np
from bs_python_utils.bsnputils import ThreeArrays, TwoArrays
from bs_python_utils.bsutils import bs_error_abort

from cupid_matching.entropy import (
    EntropyFunctions,
    EntropyHessians,
    check_additional_parameters,
)
from cupid_matching.matching_utils import Matching


def _entropy_choo_siow(
    muhat: Matching, deriv: int | None = 0
) -> (
    None
    | float
    | tuple[float, np.ndarray]
    | tuple[float, np.ndarray, np.ndarray, np.ndarray]
):
    """Returns the values of $\\mathcal{E}$
    and the first (if `deriv` is 1 or 2) and second (if `deriv` is 2) derivatives
    for the Choo and Siow model

    Args:
        muhat: a Matching
        deriv: if equal 1, we compute_ the first derivatives too;
               if equals 2, also the hessian

    Returns:
        the value of the generalized entropy
        if deriv = 1 or 2, the (X,Y) matrix of the first derivative of the entropy
        if deriv = 2, the (X,Y,X,Y) array of the second derivative
            wrt $(\\mu,\\mu)$
          and the (X,Y,X+Y) second derivatives
            wrt $(\\mu,(n,m))$
    """
    muxy, mux0, mu0y, n, m = muhat.unpack()

    logxy = np.log(muxy)
    logx0 = np.log(mux0)
    log0y = np.log(mu0y)

    val_entropy = (
        -2.0 * np.sum(muxy * logxy)
        - np.sum(mux0 * logx0)
        - np.sum(mu0y * log0y)
        + np.sum(n * np.log(n))
        + np.sum(m * np.log(m))
    )

    if deriv == 0:
        return cast(float, val_entropy)
    if deriv in [1, 2]:
        der_xy = -2.0 * logxy + log0y
        der_xy += logx0.reshape((-1, 1))
        if deriv == 1:
            return val_entropy, der_xy
        else:  # we compute_ the Hessians
            X, Y = muxy.shape
            derlogxy = 1.0 / muxy
            derlogx0 = 1.0 / mux0
            derlog0y = 1.0 / mu0y
            der2_xyzt = np.zeros((X, Y, X, Y))
            der2_xyr = np.zeros((X, Y, X + Y))
            for x in range(X):
                dlogx0 = derlogx0[x]
                for y in range(Y):
                    d2xy = np.zeros((X, Y))
                    d2xy[x, :] = -dlogx0
                    d2xy[:, y] -= derlog0y[y]
                    d2xy[x, y] -= 2.0 * derlogxy[x, y]
                    der2_xyzt[x, y, :, :] = d2xy
                    der2_xyr[x, y, x] = derlogx0[x]
                    der2_xyr[x, y, X + y] = derlog0y[y]
            return val_entropy, der_xy, der2_xyzt, der2_xyr
    else:
        bs_error_abort("deriv should be 0, 1, or 2")
        return None


def _der_entropy_choo_siow_corrected(
    muhat: Matching, hessian: bool | None = False
) -> np.ndarray | ThreeArrays:
    """Returns the corrected first derivative of $\\mathcal{E}$
    and the corrected second derivative (if `hessian` is True)
    for the Choo and Siow model

    Args:
        muhat: a Matching
        hessian: if `True`, also compute_ the hessian

    Returns:
        the (X,Y) matrix of the first derivative of the entropy
        if hessian is True, the (X,Y,X,Y) array of the second derivative
            wrt $(\\mu,\\mu)$
          and the (X,Y,X+Y) second derivatives
            wrt $(\\mu,(n,m))$
    """
    muxy, mux0, mu0y, *_ = muhat.unpack()
    n_households = np.sum(muxy) + np.sum(mux0) + np.sum(mu0y)

    muxy_corr = muxy + (1.0 - muxy / n_households) / 2.0
    logxy = np.log(muxy_corr)
    mux0_corr = mux0 + (1.0 - mux0 / n_households) / 2.0
    logx0 = np.log(mux0_corr)
    mu0y_corr = mu0y + (1.0 - mu0y / n_households) / 2.0
    log0y = np.log(mu0y_corr)

    der_xy = -2.0 * logxy + log0y
    der_xy += logx0.reshape((-1, 1))
    if not hessian:
        return der_xy
    else:  # we compute_ the Hessians
        X, Y = muxy.shape
        f_corr = 1.0 - 1.0 / n_households / 2.0
        derlogxy = f_corr / muxy_corr
        derlogx0 = f_corr / mux0_corr
        derlog0y = f_corr / mu0y_corr
        der2_xyzt = np.zeros((X, Y, X, Y))
        der2_xyr = np.zeros((X, Y, X + Y))
        for x in range(X):
            dlogx0 = derlogx0[x]
            for y in range(Y):
                d2xy = np.zeros((X, Y))
                d2xy[x, :] = -dlogx0
                d2xy[:, y] -= derlog0y[y]
                d2xy[x, y] -= 2.0 * derlogxy[x, y]
                der2_xyzt[x, y, :, :] = d2xy
                der2_xyr[x, y, x] = derlogx0[x]
                der2_xyr[x, y, X + y] = derlog0y[y]
        return der_xy, der2_xyzt, der2_xyr


def e0_fun_choo_siow(
    muhat: Matching, additional_parameters: list | None = None
) -> np.ndarray:
    """Returns the values of $e_0$ for the Choo and Siow model.

    Args:
        muhat: a Matching

    Returns:
        the (X,Y) matrix of the first derivative of the entropy
    """
    check_additional_parameters(0, additional_parameters)
    entropy_res = cast(tuple[float, np.ndarray], _entropy_choo_siow(muhat, deriv=1))
    return cast(np.ndarray, entropy_res[1])


def e0_fun_choo_siow_corrected(
    muhat: Matching, additional_parameters: list | None = None
) -> np.ndarray:
    """Returns the values of $e_0$ for the Choo and Siow model,
    using the finite-sample correction log(p+(1-p)/(2N))

    Args:
        muhat: a Matching

    Returns:
        the (X,Y) matrix of the first derivative of the entropy
    """
    check_additional_parameters(0, additional_parameters)
    e0_val_corrected = _der_entropy_choo_siow_corrected(muhat, hessian=False)
    return e0_val_corrected


def hessian_mumu_choo_siow(
    muhat: Matching, additional_parameters: list | None = None
) -> ThreeArrays:
    """Returns the derivatives of $e_0$ in $\\mu$
    for the Choo and Siow model.

    Args:
        muhat: a Matching

    Returns:
        the three components of the hessian wrt $(\\mu,\\mu)$ of the entropy
    """
    check_additional_parameters(0, additional_parameters)
    entropy_res = cast(
        tuple[float, np.ndarray, np.ndarray, np.ndarray],
        _entropy_choo_siow(muhat, deriv=2),
    )
    hessmumu = entropy_res[2]
    muxy, *_ = muhat.unpack()
    X, Y = muxy.shape
    hess_x = np.zeros((X, Y, Y))
    hess_y = np.zeros((X, Y, X))
    hess_xy = np.zeros((X, Y))
    for x in range(X):
        for y in range(Y):
            d2xy = hessmumu[x, y, :, :]
            hess_x[x, y, :] = d2xy[x, :]
            hess_y[x, y, :] = d2xy[:, y]
            hess_xy[x, y] = d2xy[x, y]
    return hess_x, hess_y, hess_xy


def hessian_mumu_choo_siow_corrected(
    muhat: Matching, additional_parameters: list | None = None
) -> ThreeArrays:
    """Returns the derivatives of $e_0$ in $\\mu$
    for the Choo and Siow model, with the small sample correction

    Args:
        muhat: a Matching

    Returns:
        the three components of the hessian wrt $(\\mu,\\mu)$ of the entropy
    """
    check_additional_parameters(0, additional_parameters)
    _, hessmumu, _ = _der_entropy_choo_siow_corrected(muhat, hessian=True)
    muxy, *_ = muhat.unpack()
    X, Y = muxy.shape
    hess_x = np.zeros((X, Y, Y))
    hess_y = np.zeros((X, Y, X))
    hess_xy = np.zeros((X, Y))
    for x in range(X):
        for y in range(Y):
            d2xy = hessmumu[x, y, :, :]
            hess_x[x, y, :] = d2xy[x, :]
            hess_y[x, y, :] = d2xy[:, y]
            hess_xy[x, y] = d2xy[x, y]
    return hess_x, hess_y, hess_xy


def hessian_mur_choo_siow(
    muhat: Matching, additional_parameters: list | None = None
) -> TwoArrays:
    """Returns the derivatives of $e_0$ in $r$
    for the Choo and Siow model.

    Args:
        muhat: a Matching

    Returns:
        the two components of the hessian wrt $(\\mu,r)$ of the entropy
    """
    check_additional_parameters(0, additional_parameters)
    entropy_res = cast(
        tuple[float, np.ndarray, np.ndarray, np.ndarray],
        _entropy_choo_siow(muhat, deriv=2),
    )
    hessmur = entropy_res[3]
    muxy, *_ = muhat.unpack()
    X, Y = muxy.shape
    hess_nx = np.zeros((X, Y))
    hess_my = np.zeros((X, Y))
    for x in range(X):
        for y in range(Y):
            d2r = hessmur[x, y, :]
            hess_nx[x, y] = d2r[x]
            hess_my[x, y] = d2r[X + y]
    return hess_nx, hess_my


def hessian_mur_choo_siow_corrected(
    muhat: Matching, additional_parameters: list | None = None
) -> TwoArrays:
    """Returns the derivatives of $e_0$ in $r$
    for the Choo and Siow model, with the small sample correction

    Args:
        muhat: a Matching

    Returns:
        the two components of the hessian wrt $(\\mu,r)$ of the entropy
    """
    check_additional_parameters(0, additional_parameters)
    _, _, hessmur = _der_entropy_choo_siow_corrected(muhat, hessian=True)
    muxy, *_ = muhat.unpack()
    X, Y = muxy.shape
    hess_nx = np.zeros((X, Y))
    hess_my = np.zeros((X, Y))
    for x in range(X):
        for y in range(Y):
            d2r = hessmur[x, y, :]
            hess_nx[x, y] = d2r[x]
            hess_my[x, y] = d2r[X + y]
    return hess_nx, hess_my


e0_derivative_choo_siow = (hessian_mumu_choo_siow, hessian_mur_choo_siow)
e0_derivative_choo_siow_corrected = (
    hessian_mumu_choo_siow_corrected,
    hessian_mur_choo_siow_corrected,
)

entropy_choo_siow = EntropyFunctions(
    e0_fun=e0_fun_choo_siow,
    hessian="provided",
    e0_derivative=cast(EntropyHessians, e0_derivative_choo_siow),
    description="Choo and Siow homoskedastic with analytic Hessian",
)

entropy_choo_siow_corrected = EntropyFunctions(
    e0_fun=e0_fun_choo_siow_corrected,
    hessian="provided",
    e0_derivative=cast(EntropyHessians, e0_derivative_choo_siow_corrected),
    description=(
        "Choo and Siow homoskedastic with analytic Hessian and finite-sample correction"
    ),
)

entropy_choo_siow_numeric = EntropyFunctions(
    e0_fun=e0_fun_choo_siow,
    description="Choo and Siow homoskedastic with numerical Hessian",
)

entropy_choo_siow_corrected_numeric = EntropyFunctions(
    e0_fun=e0_fun_choo_siow_corrected,
    description=(
        "Choo and Siow homoskedastic with numerical Hessian and finite-sample"
        " correction"
    ),
)



"""Estimates the semilinear Choo and Siow homoskedastic (2006) model
using Poisson GLM.
"""

from math import sqrt

import numpy as np
import scipy.linalg as spla
from sklearn import linear_model

from cupid_matching.matching_utils import Matching, VarianceMatching, variance_muhat
from cupid_matching.poisson_glm_utils import PoissonGLMResults, prepare_data
from cupid_matching.utils import make_XY_K_mat


def _stderrs_u(
    varcov_gamma: np.ndarray,
    n_norm: np.ndarray,
    var_muhat_norm: VarianceMatching,
    A_inv_Z: np.ndarray,
    X: int,
    Y: int,
):
    XY = X * Y
    u_std = np.zeros(X)
    var_allmus_norm = var_muhat_norm.var_allmus
    var_munm_norm = var_muhat_norm.var_munm
    var_n_norm = var_munm_norm[XY : (XY + X), XY : (XY + X)]

    ix = XY
    for x in range(X):
        n_norm_x = n_norm[x]
        A_inv_Zx = A_inv_Z[x, :]
        var_log_nx = var_n_norm[x, x] / n_norm_x / n_norm_x
        slice_x = slice(x * Y, (x + 1) * Y)
        covar_nx = var_allmus_norm[:, ix] + np.sum(var_allmus_norm[:, slice_x], 1)
        cov_a_lognx = (A_inv_Zx @ covar_nx) / n_norm_x
        ux_var = varcov_gamma[x, x] + var_log_nx + 2.0 * cov_a_lognx
        u_std[x] = sqrt(ux_var)
        ix += 1
    return u_std


def _stderrs_v(
    varcov_gamma: np.ndarray,
    m_norm: np.ndarray,
    var_muhat_norm: VarianceMatching,
    A_inv_Z: np.ndarray,
    X: int,
    Y: int,
):
    XY = X * Y
    v_std = np.zeros(Y)
    var_allmus_norm = var_muhat_norm.var_allmus
    var_munm_norm = var_muhat_norm.var_munm
    var_m_norm = var_munm_norm[(XY + X) :, (XY + X) :]

    iy, jy = X, XY + X
    for y in range(Y):
        m_norm_y = m_norm[y]
        A_inv_Zy = A_inv_Z[iy, :]
        var_log_my = var_m_norm[y, y] / m_norm_y / m_norm_y
        slice_y = slice(y, XY, Y)
        covar_b_my = var_allmus_norm[:, jy] + np.sum(var_allmus_norm[:, slice_y], 1)
        cov_b_logmy = (A_inv_Zy @ covar_b_my) / m_norm_y
        vy_var = varcov_gamma[iy, iy] + var_log_my + 2.0 * cov_b_logmy
        v_std[y] = sqrt(vy_var)
        iy += 1
        jy += 1
    return v_std


def _stderrs_u_no_singles(
    varcov_gamma: np.ndarray,
    n_norm: np.ndarray,
    var_muhat_norm: VarianceMatching,
    A_inv_Z: np.ndarray,
    X: int,
    Y: int,
):
    XY = X * Y
    var_muxy_norm = var_muhat_norm.var_allmus[:XY, :XY]
    var_munm_norm = var_muhat_norm.var_munm
    var_n_norm = var_munm_norm[XY : (XY + X), XY : (XY + X)]
    n_norm1 = n_norm[0]
    var_log_n1 = var_n_norm[0, 0] / n_norm1 / n_norm1

    u_std = np.zeros(X)
    u_std[0] = 0.0
    slice_1 = slice(0, Y)
    for x in range(1, X):
        A_inv_Zx = A_inv_Z[x, :]
        n_norm_x = n_norm[x]
        var_log_nx = var_n_norm[x, x] / n_norm_x / n_norm_x
        slice_x = slice(x * Y, (x + 1) * Y)
        covar_nx = np.sum(var_muxy_norm[:, slice_x], 1)
        covar_n1 = np.sum(var_muxy_norm[:, slice_1], 1)
        covar_nx_n1 = var_n_norm[x, 0]
        cov_a_lognx = (A_inv_Zx @ covar_nx) / n_norm_x
        cov_a_logn1 = (A_inv_Zx @ covar_n1) / n_norm_x
        cov_lognx_logn1 = covar_nx_n1 / n_norm_x / n_norm1
        ux_var = (
            varcov_gamma[x, x]
            + var_log_nx
            + var_log_n1
            + 2.0 * cov_a_lognx
            - 2.0 * cov_a_logn1
            - 2.0 * cov_lognx_logn1
        )
        u_std[x - 1] = sqrt(ux_var)
    return u_std


def _stderrs_v_no_singles(
    varcov_gamma: np.ndarray,
    m_norm: np.ndarray,
    n_norm: np.ndarray,
    var_muhat_norm: VarianceMatching,
    A_inv_Z: np.ndarray,
    X: int,
    Y: int,
):
    XY = X * Y
    var_muxy_norm = var_muhat_norm.var_allmus[:XY, :XY]
    var_munm_norm = var_muhat_norm.var_munm
    var_n_norm = var_munm_norm[XY : (XY + X), XY : (XY + X)]
    var_m_norm = var_munm_norm[(XY + X) :, (XY + X) :]
    n_norm1 = n_norm[0]
    var_log_n1 = var_n_norm[0, 0] / n_norm1 / n_norm1

    v_std = np.zeros(Y)
    iy, jy = X, XY + X
    slice_1 = slice(0, Y)
    for y in range(Y):
        A_inv_Zy = A_inv_Z[iy, :]
        m_norm_y = m_norm[y]
        var_log_my = var_m_norm[y, y] / m_norm_y / m_norm_y
        slice_y = slice(y, XY, Y)
        covar_my = np.sum(var_muxy_norm[:, slice_y], 1)
        cov_b_logmy = (A_inv_Zy @ covar_my) / m_norm_y
        covar_n1 = np.sum(var_muxy_norm[:, slice_1], 1)
        cov_b_logn1 = (A_inv_Zy @ covar_n1) / n_norm1
        covar_my_n1 = var_munm_norm[jy, XY]
        cov_logmy_logn1 = covar_my_n1 / m_norm_y / n_norm1
        vy_var = (
            varcov_gamma[iy, iy]
            + var_log_my
            + var_log_n1
            + 2.0 * cov_b_logmy
            + 2.0 * cov_b_logn1
            + 2.0 * cov_logmy_logn1
        )
        v_std[y] = sqrt(vy_var)

    return v_std


def choo_siow_poisson_glm(
    muhat: Matching,
    phi_bases: np.ndarray,
    no_singles: bool = False,
    tol: float | None = 1e-12,
    max_iter: int | None = 10000,
    verbose: int | None = 1,
) -> PoissonGLMResults:
    """Estimates the semilinear Choo and Siow homoskedastic (2006) model
        using Poisson GLM.

    Args:
        muhat: the observed Matching
        phi_bases: an (X, Y, K) array of bases
        no_singles: if True, we do not observe the singles
        tol: tolerance level for `linear_model.PoissonRegressor.fit`
        max_iter: maximum number of iterations
            for `linear_model.PoissonRegressor.fit`
        verbose: defines how much output we want (0 = least)

    Returns:
        a `PoissonGLMResults` instance

    Example:
        ```py
        n_households = 1e6
        X, Y, K = 4, 3, 6
        # we setup a quadratic set of basis functions
        phi_bases = np.zeros((X, Y, K))
        phi_bases[:, :, 0] = 1
        for x in range(X):
            phi_bases[x, :, 1] = x
            phi_bases[x, :, 3] = x * x
            for y in range(Y):
                phi_bases[x, y, 4] = x * y
        for y in range(Y):
            phi_bases[:, y, 2] = y
            phi_bases[:, y, 5] = y * y

        lambda_true = np.random.randn(K)
        phi_bases = np.random.randn(X, Y, K)
        Phi = phi_bases @ lambda_true

        # we simulate a Choo and Siow sample from a population
        #  with equal numbers of men and women of each type
        n = np.ones(X)
        m = np.ones(Y)
        choo_siow_instance = ChooSiowPrimitives(Phi, n, m)
        mus_sim = choo_siow_instance.simulate(n_households)
        muxy_sim, mux0_sim, mu0y_sim, n_sim, m_sim = mus_sim.unpack()

        results = choo_siow_poisson_glm(mus_sim, phi_bases)

        # compare true and estimated parameters
        results.print_results(
            lambda_true,
            u_true=-np.log(mux0_sim / n_sim),
            v_true=-np.log(mu0y_sim / m_sim)
        )
        ```

    """
    X, Y, K = phi_bases.shape
    XY = X * Y

    # the vector of weights for the Poisson regression
    w = (
        2 * np.ones(XY)
        if no_singles
        else np.concatenate((2 * np.ones(XY), np.ones(X + Y)))
    )
    # reshape the bases
    phi_mat = make_XY_K_mat(phi_bases)

    id_X = np.eye(X)
    id_Y = np.eye(Y)
    ones_X = np.ones((X, 1))
    ones_Y = np.ones((Y, 1))
    if no_singles:
        Z_unweighted = np.hstack(
            [-np.kron(id_X, ones_Y), -np.kron(ones_X, id_Y), phi_mat]
        )
        # we need to normalize u_1 = 0, so we delete the first column
        Z_unweighted = Z_unweighted[:, 1:]
    else:
        zeros_XK = np.zeros((X, K))
        zeros_YK = np.zeros((Y, K))
        zeros_XY = np.zeros((X, Y))
        zeros_YX = np.zeros((Y, X))
        Z_unweighted = np.vstack(
            [
                np.hstack([-np.kron(id_X, ones_Y), -np.kron(ones_X, id_Y), phi_mat]),
                np.hstack([-id_X, zeros_XY, zeros_XK]),
                np.hstack([zeros_YX, -id_Y, zeros_YK]),
            ]
        )
    Z = Z_unweighted / w.reshape((-1, 1))

    var_muhat = variance_muhat(muhat)
    (
        muhat_norm,
        var_muhat_norm,
        n_households,
        n_individuals,
    ) = prepare_data(muhat, var_muhat, no_singles=no_singles)

    clf = linear_model.PoissonRegressor(
        fit_intercept=False,
        tol=tol,
        verbose=verbose,
        alpha=0,
        max_iter=max_iter,
    )
    if no_singles:
        muxyhat_norm = muhat_norm[:XY]
        clf.fit(Z, muxyhat_norm, sample_weight=w)
    else:
        clf.fit(Z, muhat_norm, sample_weight=w)
    gamma_est = clf.coef_

    # we compute_ the variance-covariance of the estimator
    var_allmus_norm = var_muhat_norm.var_allmus
    var_norm = var_allmus_norm[:XY, :XY] if no_singles else var_allmus_norm
    nr, nc = Z.shape
    exp_Zg = np.exp(Z @ gamma_est).reshape(nr)
    A_hat = np.zeros((nc, nc))
    B_hat = np.zeros((nc, nc))
    for i in range(nr):
        Zi = Z[i, :]
        wi = w[i]
        A_hat += wi * exp_Zg[i] * np.outer(Zi, Zi)
        for j in range(nr):
            Zj = Z[j, :]
            B_hat += wi * w[j] * var_norm[i, j] * np.outer(Zi, Zj)

    A_inv = spla.inv(A_hat)
    varcov_gamma = A_inv @ B_hat @ A_inv
    stderrs_gamma = np.sqrt(np.diag(varcov_gamma))

    beta_est = gamma_est[-K:]
    varcov_beta = varcov_gamma[-K:, -K:]
    beta_std = stderrs_gamma[-K:]
    Phi_est = phi_bases @ beta_est

    # we correct for the effect of the normalization
    _, _, _, n, m = muhat.unpack()
    n_norm = n / n_individuals
    m_norm = m / n_individuals
    if no_singles:
        u_est = gamma_est[: (X - 1)]
        v_est = gamma_est[(X - 1) : -K]
        # normalize u_1 = 0
        n_0 = n_norm[0]
        u_est = np.concatenate((np.zeros(1), u_est + np.log(n_norm[1:] / n_0)))
        v_est += np.log(m_norm * n_0)
    else:
        u_est = gamma_est[:X] + np.log(n_norm)
        v_est = gamma_est[X:-K] + np.log(m_norm)

    # since u and v are translated from gamma we need to adjust the estimated stderrs
    A_inv_Z = A_inv @ Z_unweighted.T
    if no_singles:
        u_std = _stderrs_u_no_singles(
            varcov_gamma, n_norm, var_muhat_norm, A_inv_Z, X, Y
        )
        v_std = _stderrs_v_no_singles(
            varcov_gamma, m_norm, n_norm, var_muhat_norm, A_inv_Z, X, Y
        )
    else:
        u_std = _stderrs_u(varcov_gamma, n_norm, var_muhat_norm, A_inv_Z, X, Y)
        v_std = _stderrs_v(varcov_gamma, m_norm, var_muhat_norm, A_inv_Z, X, Y)

    results = PoissonGLMResults(
        X=X,
        Y=Y,
        K=K,
        number_households=n_households,
        number_individuals=n_individuals,
        estimated_gamma=gamma_est,
        estimated_Phi=Phi_est,
        estimated_beta=beta_est,
        estimated_u=u_est,
        estimated_v=v_est,
        varcov_gamma=varcov_gamma,
        varcov_beta=varcov_beta,
        stderrs_gamma=stderrs_gamma,
        stderrs_beta=beta_std,
        stderrs_u=u_std,
        stderrs_v=v_std,
    )

    return results



"""The components of the derivative of the entropy
for the Choo and Siow gender-heteroskedastic model.

We normalize the standard error for the X side at 1,
and we estimate the standard error on the Y side.
"""
from typing import cast

import numpy as np
from bs_python_utils.bsnputils import ThreeArrays, TwoArrays

from cupid_matching.entropy import EntropyFunctions, EntropyHessians
from cupid_matching.matching_utils import Matching


def e0_choo_siow_gender_heteroskedastic(muhat: Matching) -> np.ndarray:
    """Returns the values of the parameter-independent part $e_0$
    for the Choo and Siow gender-heteroskedastic model; we normalized $\\sigma=1$.

    Args:
        muhat: a Matching

    Returns:
        the (X,Y) matrix of the parameter-independent part
        of the first derivative of the entropy.
    """
    muxy, mux0, *_ = muhat.unpack()
    e0_vals = -np.log(muxy / mux0.reshape((-1, 1)))
    return cast(np.ndarray, e0_vals)


def e0_derivative_mu_gender_heteroskedastic(
    muhat: Matching,
) -> ThreeArrays:
    """Returns the derivatives of the parameter-independent part $e_0$ in $\\mu$.
    for the Choo and Siow gender-heteroskedastic model; we normalized $\\sigma=1$.

    Args:
        muhat: a Matching

    Returns:
        the  parameter-independent part of the hessian of the entropy
        wrt $(\\mu,\\mu)$.
    """
    muxy, mux0, *_ = muhat.unpack()
    X, Y = muxy.shape
    hess_x = np.zeros((X, Y, Y))
    hess_y = np.zeros((X, Y, X))
    hess_xy = np.zeros((X, Y))
    der_logxy = 1.0 / muxy
    der_logx0 = 1.0 / mux0
    for x in range(X):
        dlogx0 = der_logx0[x]
        for y in range(Y):
            hess_x[x, y, :] = -dlogx0
            hess_xy[x, y] = -der_logxy[x, y] - dlogx0
    return hess_x, hess_y, hess_xy


def e0_derivative_r_gender_heteroskedastic(
    muhat: Matching,
) -> TwoArrays:
    """Returns the derivatives of the parameter-independent part $e_0$ wrt $r$
    for the Choo and Siow gender-heteroskedastic model; we normalized $\\sigma=1$.

    Args:
        muhat: a Matching

    Returns:
        the parameter-independent part of the hessian of the entropy
        wrt $(\\mu,r)$.
    """
    muxy, mux0, *_ = muhat.unpack()
    X, Y = muxy.shape
    hess_n = np.zeros((X, Y))
    hess_m = np.zeros((X, Y))
    der_logx0 = 1.0 / mux0
    for x in range(X):
        dlogx0 = der_logx0[x]
        for y in range(Y):
            hess_n[x, y] = dlogx0
    return hess_n, hess_m


e0_derivative_choo_siow_gender_heteroskedastic = (
    e0_derivative_mu_gender_heteroskedastic,
    e0_derivative_r_gender_heteroskedastic,
)


def e_choo_siow_gender_heteroskedastic(muhat: Matching) -> np.ndarray:
    """Returns the values of the parameter-dependent part  $e$
    for the Choo and Siow gender-heteroskedastic model; we normalized $\\sigma=1$.

    Args:
        muhat: a Matching

    Returns:
        the (X,Y,1) array of the parameter-dependent part
        of the first derivative of the entropy.
    """
    muxy, _, mu0y, *_ = muhat.unpack()
    X, Y = muxy.shape
    n_alpha = 1

    e_vals = np.zeros((X, Y, n_alpha))
    e_vals[:, :, 0] = -np.log(muxy / mu0y)
    return e_vals


def e_derivative_mu_gender_heteroskedastic(
    muhat: Matching,
) -> ThreeArrays:
    """Returns the derivatives of the parameter-dependent part $e$
     wrt $\\mu$ for the Choo and Siow gender-heteroskedastic model;
     we normalized $\\sigma_1=1$.

    Args:
        muhat: a Matching

    Returns:
        the parameter-dependent part of the hessian of the entropy
        wrt $(\\mu,\\mu)$.
    """
    muxy, _, mu0y, *_ = muhat.unpack()
    X, Y = muxy.shape

    n_alpha = 1
    hess_x = np.zeros((X, Y, Y, n_alpha))
    hess_y = np.zeros((X, Y, X, n_alpha))
    hess_xy = np.zeros((X, Y, n_alpha))
    der_logxy = 1.0 / muxy
    der_log0y = 1.0 / mu0y
    for x in range(X):
        for y in range(Y):
            dlog0y = der_log0y[y]
            hess_y[x, y, :, 0] = -dlog0y
            hess_xy[x, y, 0] = -der_logxy[x, y] - dlog0y
    return hess_x, hess_y, hess_xy


def e_derivative_r_gender_heteroskedastic(
    muhat: Matching,
) -> TwoArrays:
    """Returns the derivatives of the parameter-dependent part $e$
     wrt $r$ for the Choo and Siow gender-heteroskedastic model;
     we normalized $\\sigma_1=1$.

    Args:
        muhat: a Matching

    Returns:
        the parameter-dependent part of the hessian of the entropy
        wrt $(\\mu,r)$
    """
    muxy, _, mu0y, *_ = muhat.unpack()
    X, Y = muxy.shape

    n_alpha = 1
    hess_n = np.zeros((X, Y, n_alpha))
    hess_m = np.zeros((X, Y, n_alpha))
    der_log0y = 1.0 / mu0y
    for x in range(X):
        for y in range(Y):
            dlog0y = der_log0y[y]
            hess_m[x, y, 0] = dlog0y
    return hess_n, hess_m


e_derivative_choo_siow_gender_heteroskedastic = (
    e_derivative_mu_gender_heteroskedastic,
    e_derivative_r_gender_heteroskedastic,
)

entropy_choo_siow_gender_heteroskedastic = EntropyFunctions(
    e0_fun=e0_choo_siow_gender_heteroskedastic,
    parameter_dependent=True,
    e_fun=e_choo_siow_gender_heteroskedastic,
    hessian="provided",
    e0_derivative=cast(EntropyHessians, e0_derivative_choo_siow_gender_heteroskedastic),
    e_derivative=cast(EntropyHessians, e_derivative_choo_siow_gender_heteroskedastic),
    description="Choo and Siow gender-heteroskedastic with analytic Hessian",
)

entropy_choo_siow_gender_heteroskedastic_numeric = EntropyFunctions(
    e0_fun=e0_choo_siow_gender_heteroskedastic,
    parameter_dependent=True,
    e_fun=e_choo_siow_gender_heteroskedastic,
    description="Choo and Siow gender-heteroskedastic with numerical Hessian",
)



""" matching-related utilities """

from dataclasses import dataclass, field
from typing import Any, Final, Protocol, cast

import numpy as np
from bs_python_utils.bsnputils import TwoArrays, check_matrix, check_vector, npmaxabs
from bs_python_utils.bsutils import bs_error_abort

SINGLES_TOL: Final = 1e-3


def get_singles(muxy: np.ndarray, n: np.ndarray, m: np.ndarray) -> TwoArrays:
    """Computes the numbers of singles from the matches and the margins."""
    mux0 = n - np.sum(muxy, 1)
    mu0y = m - np.sum(muxy, 0)
    return mux0, mu0y


def compute_margins(muxy: np.ndarray, mux0: np.ndarray, mu0y: np.ndarray) -> TwoArrays:
    """Computes the margins from the matches and the singles."""
    n = np.sum(muxy, 1) + mux0
    m = np.sum(muxy, 0) + mu0y
    return n, m


def _check_no_singles(mu0: np.ndarray, n_households: float, str_gender: str) -> None:
    maxabs = npmaxabs(mu0)
    if maxabs > SINGLES_TOL * n_households:
        bs_error_abort(
            f"In a model w/o singles, we should not have any single {str_gender}."
        )


@dataclass
class Matching:
    """stores the numbers of couples and singles of every type;

    `muxy` is an (X,Y)-matrix
    `n` is an X-vector
    `m` is an Y-vector

    `no_singles`: if `True`, this is a model w/o singles

    `mux0` and `mu0y` are generated as the corresponding numbers of singles
    as well as the total number of households `n_households`
    and the total number of individuals `n_individuals`
    """

    muxy: np.ndarray
    n: np.ndarray
    m: np.ndarray
    no_singles: bool = False

    mux0: np.ndarray = field(init=False)
    mu0y: np.ndarray = field(init=False)
    n_households: float = field(init=False)
    n_individuals: float = field(init=False)

    def __str__(self):
        X, Y = self.muxy.shape
        n_couples = np.sum(self.muxy)
        n_men, n_women = np.sum(self.n), np.sum(self.m)
        repr_str = f"This is a matching with {n_men} men and {n_women} women.\n"
        repr_str += f"   with {n_couples} couples,\n"
        if self.no_singles:
            repr_str += "     and no singles.\n"
        repr_str += f"\n We have {X} types of men and {Y} of women."
        return repr_str

    def __post_init__(self):
        X, Y = check_matrix(self.muxy)
        Xn = check_vector(self.n)
        Ym = check_vector(self.m)
        if Xn != X:
            bs_error_abort(f"muxy is a ({X}, {Y}) matrix but n has {Xn} elements.")
        if Ym != Y:
            bs_error_abort(f"muxy is a ({X}, {Y}) matrix but m has {Ym} elements.")
        self.mux0, self.mu0y = get_singles(self.muxy, self.n, self.m)
        self.n_households = np.sum(self.muxy) + np.sum(self.mux0) + np.sum(self.mu0y)
        if self.no_singles:
            _check_no_singles(self.mux0, self.n_households, "men")
            _check_no_singles(self.mu0y, self.n_households, "women")
        self.n_individuals = (
            2.0 * np.sum(self.muxy) + np.sum(self.mux0) + np.sum(self.mu0y)
        )

    def unpack(self):
        muxy, mux0, mu0y = self.muxy, self.mux0, self.mu0y
        min_xy, min_x0, min_0y = np.min(muxy), np.min(mux0), np.min(mu0y)
        if min_xy < 0.0:
            bs_error_abort(f"The smallest muxy is {min_xy}")
        if not self.no_singles:
            if min_x0 < 0.0:
                bs_error_abort(f"The smallest mux0 is {min_x0}")
            if min_0y < 0.0:
                bs_error_abort(f"The smallest mux0 is {min_0y}")
        return muxy, mux0, mu0y, self.n, self.m


class MatchingFunction(Protocol):
    def __call__(
        self, mus: Matching, additional_parameters: list | None = ..., /
    ) -> np.ndarray:
        ...


def get_evals(
    fun: MatchingFunction, mus: Matching, additional_parameters: list | None = None
) -> np.ndarray:
    """evaluates fun(mus, additional_parameters)"""
    vals = fun(mus, additional_parameters)
    return cast(np.ndarray, vals)


def get_margins(mus: Matching) -> TwoArrays:
    """computes the numbers of each type from the matching patterns"""
    _, _, _, n, m = mus.unpack()
    return n, m


def simulate_sample_from_mus(
    mus: Matching, n_households: int, no_singles: bool = False, seed: int | None = None
) -> Matching:
    """Draw a sample of `n_households` from the matching patterns in `mus`

    Args:
        mus: the matching patterns
        n_households: the number of households requested
        no_singles: if `True`, this is a model w/o singles
        seed: an integer seed for the random number generator

    Returns:
        the sample matching patterns
    """
    rng = np.random.default_rng(seed)
    muxy, mux0, mu0y, _, _ = mus.unpack()
    X, Y = muxy.shape
    # stack all probabilities
    XY = X * Y
    # make sure we have no zeros
    _MU_EPS = min(1, int(1e-3 * n_households))
    if no_singles:
        pvec = muxy.reshape(XY)
        pvec /= np.sum(pvec)
        matches = rng.multinomial(n_households, pvec)
        muxy_sim = matches.reshape((X, Y))
        mux0_sim = np.full(X, _MU_EPS)
        mu0y_sim = np.full(Y, _MU_EPS)
    else:
        num_choices = XY + X + Y
        pvec = np.zeros(num_choices)
        pvec[:XY] = muxy.reshape(XY)
        pvec[XY : (XY + X)] = mux0
        pvec[(XY + X) :] = mu0y
        pvec /= np.sum(pvec)
        matches = rng.multinomial(n_households, pvec)
        muxy_sim = matches[:XY].reshape((X, Y))
        mux0_sim = matches[XY : (XY + X)]
        mu0y_sim = matches[(XY + X) :]
        muxy_sim += _MU_EPS
        mux0_sim += _MU_EPS
        mu0y_sim += _MU_EPS
    n_sim, m_sim = compute_margins(muxy_sim, mux0_sim, mu0y_sim)
    mus_sim = Matching(muxy=muxy_sim, n=n_sim, m=m_sim, no_singles=no_singles)
    return mus_sim


@dataclass
class VarianceMatching:
    """initialized with the six matrix components of the variance of a `Matching;  computes five more components.

    `var_xyzt` is the (XY, XY) var-cov matrix of `muxy`
    `var_xyz0` is the (XY, X) covariance matrix of `muxy` and `mux0`
    `var_xy0t` is the (XY, Y) covariance matrix of `muxy` and `mu0y`
    `var_x0z0` is the (X, X) var-cov matrix of `mux0`
    `var_x00t` is the (X, Y) covariance matrix of `mux0` and `mu0y`
    `var_0y0t` is the (Y, Y) var-cov matrix of `mu0y`

    `var_xyn` is the (XY, X) covariance matrix of `muxy` and `nx`
    `var_xym` is the (XY, Y) covariance matrix of `muxy` and `my`
    `var_nn` is the (X, X) var-cov matrix of `nx`
    `var_nm` is the (X, Y) covariance matrix of `nx` and `my`
    `var_mm` is the (Y, Y) var-cov matrix of `my`

    `var_allmus` is the (XY+X+Y, XY+X+Y) var-cov matrix of (muxy, mux0, mu0y)
    `var_munm` is the (XY+X+Y, XY+X+Y) var-cov matrix of (muxy, n, m)

    `no_singles`: if `True`, this is a model w/o singles;
        then `var_allmus` and `var_munm` are `(XY, XY)` and `(XY, XY+X+Y)`matrices
    """

    var_xyzt: np.ndarray
    var_xyz0: np.ndarray
    var_xy0t: np.ndarray
    var_x0z0: np.ndarray
    var_x00t: np.ndarray
    var_0y0t: np.ndarray

    no_singles: bool = False

    var_xyn: np.ndarray = field(init=False)
    var_xym: np.ndarray = field(init=False)
    var_nn: np.ndarray = field(init=False)
    var_nm: np.ndarray = field(init=False)
    var_mm: np.ndarray = field(init=False)

    var_allmus: np.ndarray = field(init=False)
    var_munm: np.ndarray = field(init=False)

    def __str__(self):
        n_men = self.var_xyz0.shape[1]
        n_women = self.var_xy0t.shape[1]
        repr_str = f"This is a VarianceMatching with {n_men}  men, {n_women} women.\n"
        if self.no_singles:
            repr_str += "    we have  no singles.\n\n"
        return repr_str

    def __post_init__(self):
        v_xyzt = self.var_xyzt
        XY, XY2 = check_matrix(v_xyzt)
        if XY2 != XY:
            bs_error_abort(f"var_xyzt should be a square matrix, not ({XY}, {XY2})")
        v_xyz0 = self.var_xyz0
        XY3, X = check_matrix(v_xyz0)
        if XY3 != XY:
            bs_error_abort(f"var_xyz0 should have {XY} rows, not {XY3})")
        v_xy0t = self.var_xy0t
        XY4, Y = check_matrix(v_xy0t)
        if XY4 != XY:
            bs_error_abort(f"var_xy0t should have {XY} rows, not {XY4})")
        if X * Y != XY:
            bs_error_abort(
                f"var_xyzt has {XY} rows, but varxyz0 has {X} columns and varxy0t"
                f" has {Y}"
            )
        v_x0z0 = self.var_x0z0
        X2, X3 = check_matrix(v_x0z0)
        if X2 != X:
            bs_error_abort(f"var_x0z0 has {X2} rows, it should have {X}")
        if X3 != X:
            bs_error_abort(f"var_x0z0 has {X3} columns, it should have {X}")
        v_x00t = self.var_x00t
        X4, Y2 = check_matrix(v_x00t)
        if X4 != X:
            bs_error_abort(f"var_x00t has {X4} rows, it should have {X}")
        if Y2 != Y:
            bs_error_abort(f"var_x00t has {Y2} columns, it should have {Y}")
        v_0y0t = self.var_0y0t
        Y3, Y4 = check_matrix(v_0y0t)
        if Y3 != Y:
            bs_error_abort(f"var_x00t has {Y3} rows, it should have {Y}")
        if Y4 != Y:
            bs_error_abort(f"var_x00t has {Y4} columns, it should have {Y}")

        # now we compute the additional components
        if self.no_singles:
            v_xyn = np.zeros_like(v_xyz0)
            iz = 0
            for z in range(X):
                v_xyn[:, z] += np.sum(v_xyzt[:, iz : (iz + Y)], 1)
                iz += Y
            self.var_xyn = v_xyn
            v_xym = np.zeros_like(v_xy0t)
            for t in range(Y):
                slice_t = slice(t, XY, Y)
                v_xym[:, t] += np.sum(v_xyzt[:, slice_t], 1)
            self.var_xym = v_xym
            v_nn = np.zeros_like(v_x0z0)
            ix = 0
            for x in range(X):
                v_nn[x, :] += np.sum(v_xyn[ix : (ix + Y), :], 0)
                ix += Y
            self.var_nn = v_nn
            v_nm = np.zeros_like(v_x00t)
            v_mm = np.zeros_like(v_0y0t)
            for y in range(Y):
                slice_y = slice(y, XY, Y)
                v_nm[:, y] += np.sum(v_xyn[slice_y, :], 0)
                v_mm[y, :] += np.sum(v_xym[slice_y, :], 0)
            self.var_nm = v_nm
            self.var_mm = v_mm
        else:
            v_xyn = v_xyz0.copy()
            sumt_covx0_zt = np.zeros((X, X))
            iz = 0
            for z in range(X):
                v_xyn[:, z] += np.sum(v_xyzt[:, iz : (iz + Y)], 1)
                sumt_covx0_zt[:, z] = np.sum(v_xyz0[iz : (iz + Y), :], 0)
                iz += Y
            self.var_xyn = v_xyn
            v_xym = v_xy0t.copy()
            v_0ym = v_0y0t.copy()
            for t in range(Y):
                slice_t = slice(t, XY, Y)
                v_xym[:, t] += np.sum(v_xyzt[:, slice_t], 1)
                v_0ym[:, t] += np.sum(v_xy0t[slice_t, :], 0)
            self.var_xym = v_xym
            v_x0n = v_x0z0 + sumt_covx0_zt
            v_nn = v_x0n
            v_0yn = v_x00t.copy().T
            ix = 0
            for x in range(X):
                v_nn[x, :] += np.sum(v_xyn[ix : (ix + Y), :], 0)
                v_0yn[:, x] += np.sum(v_xy0t[ix : (ix + Y), :], 0)
                ix += Y
            self.var_nn = v_nn
            v_nm = v_0yn.T
            v_mm = v_0ym
            for y in range(Y):
                slice_y = slice(y, XY, Y)
                v_nm[:, y] += np.sum(v_xyn[slice_y, :], 0)
                v_mm[y, :] += np.sum(v_xym[slice_y, :], 0)
            self.var_nm = v_nm
            self.var_mm = v_mm

        self.var_allmus = self.make_var_allmus()
        self.var_munm = self.make_var_munm()

    def unpack(self):
        """return a tuple of all members of this `VarianceMatching`"""
        return (
            self.var_xyzt,
            self.var_xyz0,
            self.var_xy0t,
            self.var_x0z0,
            self.var_x00t,
            self.var_0y0t,
            self.var_xyn,
            self.var_xym,
            self.var_nn,
            self.var_nm,
            self.var_mm,
        )

    def make_var_allmus(self: Any) -> np.ndarray:
        """create the variance-covariance of `(muxy, mux0, mu0y)`

        Args:
            self:  the `VarianceMatching` object

        Returns:
            an `(XY+X+Y, XY+X+Y)` symmetric positive matrix if there are singles; otherwise `(XY, XY)`
        """
        v_xyzt, v_xyz0, v_xy0t, v_x0z0, v_x00t, v_0y0t, *_ = self.unpack()
        X, Y = v_x0z0.shape[0], v_0y0t.shape[0]
        XY = X * Y

        if not self.no_singles:
            sz = XY + X + Y
            v_allmus = np.zeros((sz, sz))
            v_allmus[:XY, :XY] = v_xyzt
            v_allmus[:XY, XY : (XY + X)] = v_xyz0
            v_allmus[XY : (XY + X), :XY] = v_xyz0.T
            v_allmus[:XY, (XY + X) :] = v_xy0t
            v_allmus[(XY + X) :, :XY] = v_xy0t.T
            v_allmus[XY : (XY + X), XY : (XY + X)] = v_x0z0
            v_allmus[XY : (XY + X), (XY + X) :] = v_x00t
            v_allmus[(XY + X) :, XY : (XY + X)] = v_x00t.T
            v_allmus[(XY + X) :, (XY + X) :] = v_0y0t
        else:
            v_allmus = v_xyzt

        return v_allmus

    def make_var_munm(self: Any) -> np.ndarray:
        """create the variance-covariance of `(muxy, n, m)`

        Args:
            self:  this `VarianceMatching` object

        Returns:
            an `(XY+X+Y, XY+X+Y)` symmetric positive matrix.
        """
        v_xyzt, *_, v_xyn, v_xym, v_nn, v_nm, v_mm = self.unpack()
        X, Y = v_nn.shape[0], v_mm.shape[0]
        XY = X * Y
        sz = XY + X + Y
        v_munm = np.zeros((sz, sz))
        v_munm[:XY, :XY] = v_xyzt
        v_munm[:XY, XY : (XY + X)] = v_xyn
        v_munm[XY : (XY + X), :XY] = v_xyn.T
        v_munm[:XY, (XY + X) :] = v_xym
        v_munm[(XY + X) :, :XY] = v_xym.T
        v_munm[XY : (XY + X), XY : (XY + X)] = v_nn
        v_munm[XY : (XY + X), (XY + X) :] = v_nm
        v_munm[(XY + X) :, XY : (XY + X)] = v_nm.T
        v_munm[(XY + X) :, (XY + X) :] = v_mm

        return v_munm


def var_divide(varmus: VarianceMatching, d: float) -> VarianceMatching:
    """divide all members by the same number"""
    vardiv = VarianceMatching(
        varmus.var_xyzt / d,
        varmus.var_xyz0 / d,
        varmus.var_xy0t / d,
        varmus.var_x0z0 / d,
        varmus.var_x00t / d,
        varmus.var_0y0t / d,
    )
    return vardiv


def variance_muhat(muhat: Matching) -> VarianceMatching:
    """
    Computes the unweighted variance-covariance matrix of the observed matching patterns

    Args:
        muhat: a `Matching` object

    Returns:
        the corresponding `VarianceMatching` object
    """
    muxy, mux0, mu0y, *_ = muhat.unpack()
    X, Y = muxy.shape
    XY = X * Y

    # normalize all proportions
    n_households = muhat.n_households
    muxy_norm = (muxy / n_households).ravel()
    mux0_norm = mux0 / n_households
    mu0y_norm = mu0y / n_households

    # we construct the variance of (muxy, mux0, mu0y)
    # variance of muxy
    v_xyzt = np.diag(muxy_norm) - np.outer(muxy_norm, muxy_norm)
    if muhat.no_singles:
        v_xyz0 = np.zeros((XY, X))
        v_xy0t = np.zeros((XY, Y))
        v_x0z0 = np.zeros((X, X))
        v_x00t = np.zeros((X, Y))
        v_0y0t = np.zeros((Y, Y))
    else:
        # covariance of muxy and mux0
        v_xyz0 = -np.outer(muxy_norm, mux0_norm)
        # covariance of muxy and mu0y
        v_xy0t = -np.outer(muxy_norm, mu0y_norm)
        # variance of mux0
        v_x0z0 = np.diag(mux0_norm) - np.outer(mux0_norm, mux0_norm)
        # covariance of mux0 and mu0y
        v_x00t = -np.outer(mux0_norm, mu0y_norm)
        # variance of mu0y
        v_0y0t = np.diag(mu0y_norm) - np.outer(mu0y_norm, mu0y_norm)

    v_xyzt *= n_households
    v_xyz0 *= n_households
    v_xy0t *= n_households
    v_x0z0 *= n_households
    v_x00t *= n_households
    v_0y0t *= n_households

    varmus = VarianceMatching(
        var_xyzt=v_xyzt,
        var_xyz0=v_xyz0,
        var_xy0t=v_xy0t,
        var_x0z0=v_x0z0,
        var_x00t=v_x00t,
        var_0y0t=v_0y0t,
    )
    return varmus



"""Utility programs used in `min_distance.py`.
"""

from dataclasses import dataclass
from typing import cast

import numpy as np
import scipy.linalg as spla
from bs_python_utils.bsnputils import ThreeArrays, TwoArrays, npmaxabs
from bs_python_utils.bsutils import bs_error_abort, print_stars

from cupid_matching.entropy import (
    fill_hessianMuMu_from_components,
    fill_hessianMuR_from_components,
)
from cupid_matching.matching_utils import Matching, variance_muhat


def check_args_mde(muhat: Matching, phi_bases: np.ndarray) -> tuple[int, int, int]:
    """check that the arguments to the MDE are consistent"""
    muxyhat, *_ = muhat.unpack()
    X, Y = muxyhat.shape
    ndims_phi = phi_bases.ndim
    if ndims_phi != 3:
        bs_error_abort(f"phi_bases should have 3 dimensions, not {ndims_phi}")
    Xp, Yp, K = phi_bases.shape
    if Xp != X or Yp != Y:
        bs_error_abort(
            f"phi_bases should have shape ({X}, {Y}, {K}) not ({Xp}, {Yp}, {K})"
        )
    return X, Y, K


def get_initial_weighting_matrix(
    parameterized_entropy: bool, initial_weighting_matrix: np.ndarray | None, XY: int
) -> np.ndarray | None:
    """returns the initial weighting matrix for the MDE when the entropy is parameterized

    Args:
        parameterized_entropy: if `True`, the entropy has unknown parameters
        initial_weighting_matrix: the initial weighting matrix, if provided
        XY: = X*Y

    Returns:
        the initial_weighting_matrix, or None if the entropy is not parameterized.
    """
    if parameterized_entropy:
        if initial_weighting_matrix is None:
            print_stars(
                "Using the identity matrix as weighting matrix in the first step."
            )
            S_mat = np.eye(XY)
        else:
            S_mat = initial_weighting_matrix
        return S_mat
    else:
        return None


def make_D2_matrix(X: int, Y: int) -> tuple[np.ndarray, int]:
    """create the double difference matrix for use w/o singles

    Args:
        X: number of types of men
        Y: number of types of women

    Returns:
        an (r, XY) matrix and  its rank r
    """
    XY = X * Y
    D2_mat = np.ones((XY, XY)) / XY + np.eye(XY)
    for x in range(X):
        slice_x = slice(x * Y, x * Y + Y)
        D2_mat[slice_x, slice_x] -= 1.0 / Y
    for y in range(Y):
        slice_y = slice(y, XY, Y)
        D2_mat[slice_y, slice_y] -= 1.0 / X
    rank_D2 = np.linalg.matrix_rank(D2_mat)
    rng = np.random.default_rng(453)
    A_mat = rng.uniform(size=(rank_D2, XY))
    D2_mat = A_mat @ D2_mat
    rank_D2 = np.linalg.matrix_rank(D2_mat)
    print(f"\nThe rank of the double difference matrix D2 is {rank_D2}.\n")
    return D2_mat, rank_D2


def check_indep_phi_no_singles(D2_phi: np.ndarray, X: int, Y: int) -> None:
    """check that the double difference of the phi matrix has full column rank;
        if so, return it

    Args:
        D2_phi: an $(X*Y, K)$ matrix of double differences
        X: number of types of men
        Y: number of types of women

    Returns:
        nothing
    """
    K = D2_phi.shape[1]
    actual_rank = np.linalg.matrix_rank(D2_phi)  # Compute the matrix rank
    if actual_rank != D2_phi.shape[1]:
        bs_error_abort(
            f"We have {K} basis functions but phi_mat only has rank {actual_rank}."
        )


def make_hessian_mde(
    hessian_components_mumu: ThreeArrays, hessian_components_mur: TwoArrays
) -> np.ndarray:
    """reconstitute the Hessian of the entropy function from its components

    Args:
        hessian_components_mumu:  the components of the Hesssian wrt $(\\mu,\\mu)$
        hessian_components_mur: the components of the Hesssian wrt $(\\mu,r)$

    Returns:
        np.ndarray: _description_
    """
    hessian_mumu = fill_hessianMuMu_from_components(hessian_components_mumu)
    hessian_mur = fill_hessianMuR_from_components(hessian_components_mur)
    hessians_both = np.concatenate((hessian_mumu, hessian_mur), axis=1)
    return cast(np.ndarray, hessians_both)


def get_optimal_weighting_matrix(
    muhat: Matching,
    hessians_both: np.ndarray,
    no_singles: bool = False,
    D2_mat: np.ndarray | None = None,
) -> np.ndarray:
    """compute the $S^\ast$ matrix used in the second step of the MDE

    Args:
        muhat: the observed `Matching`
        hessians_both: the Hessian of the entropy function
    """
    var_muhat = variance_muhat(muhat)
    var_munm = var_muhat.var_munm
    var_entropy_gradient = hessians_both @ var_munm @ hessians_both.T
    if no_singles:
        if D2_mat is None:
            bs_error_abort("D2_mat should not be None when no_singles is True")
        else:
            var_entropy_gradient = D2_mat @ var_entropy_gradient @ D2_mat.T
    S_mat = spla.inv(var_entropy_gradient)
    return cast(np.ndarray, S_mat)


def compute_estimates(
    M: np.ndarray, S_mat: np.ndarray, d: np.ndarray
) -> tuple[np.ndarray, np.ndarray]:
    """Returns the QGLS estimates and their variance-covariance.

    Args:
        M: an (XY,p) matrix
        S_mat: an (XY, XY) weighting matrix
        d: an XY-vector

    Returns:
        the p-vector of estimates and their estimated (p,p) variance
    """
    M_T = M.T
    M_S_d = M_T @ S_mat @ d
    M_S_M = M_T @ S_mat @ M
    est_coeffs = -spla.solve(M_S_M, M_S_d)
    varcov_coeffs = spla.inv(M_S_M)
    return est_coeffs, varcov_coeffs


@dataclass
class MDEResults:
    """
    The results from minimum-distance estimation and testing.

    Args:
        X: the number of types of men
        Y: the number of types of women
        K: the number of bases
        number_households: the number of households in the sample
        estimated_coefficients: the estimated coefficients
        varcov_coefficients: their eetimated var-covar
        stderrs_coefficients: their estimated stderrs
        estimated_Phi: the estimated joint surplus
        test_statistic: the value of the misspecification statistic
        test_pvalue: the p-value of the test
        ndf: the number of degrees of freedom
        parameterized_entropy: True if the derivative of the entropy has unknown parameters
    """

    X: int
    Y: int
    K: int
    number_households: int
    estimated_coefficients: np.ndarray
    varcov_coefficients: np.ndarray
    stderrs_coefficients: np.ndarray
    estimated_Phi: np.ndarray
    test_statistic: float
    test_pvalue: float
    ndf: int
    parameterized_entropy: bool | None = False

    def __str__(self):
        line_stars = "*" * 80 + "\n"
        if self.parameterized_entropy:
            n_alpha = self.estimated_coefficients.size - self.K
            entropy_str = f"     The entropy has {n_alpha} parameters."
        else:
            entropy_str = "     The entropy is parameter-free."
            n_alpha = 0
        model_str = f"The data has {self.number_households} households\n\n"
        model_str += f"The model has {self.X}x{self.Y} margins\n {entropy_str} \n"
        model_str += f"We use {self.K} basis functions.\n\n"
        repr_str = line_stars + model_str
        repr_str += "The estimated coefficients (and their standard errors) are\n\n"
        if self.parameterized_entropy:
            for i, coeff in enumerate(self.estimated_coefficients[:n_alpha]):
                repr_str += (
                    f"   alpha({i + 1}): {coeff: > 10.3f}  "
                    + f"({self.stderrs_coefficients[i]: .3f})\n"
                )
            repr_str += "\n"
        for i, coeff in enumerate(self.estimated_coefficients[n_alpha:]):
            repr_str += (
                f"   base {i + 1}: {coeff: > 10.3f} "
                + f"({self.stderrs_coefficients[n_alpha + i]: .3f})\n"
            )
        repr_str += "\nSpecification test:\n"
        repr_str += (
            f"   the value of the test statistic is {self.test_statistic: > 10.3f}\n"
        )
        repr_str += (
            f"     for a chi2({self.ndf}), the p-value is {self.test_pvalue: > 10.3f}\n"
        )
        return repr_str + line_stars

    def print_results(
        self, true_coeffs: np.ndarray | None = None, n_alpha: int = 0
    ) -> None | float:
        estimates = self.estimated_coefficients
        stderrs = self.stderrs_coefficients

        if true_coeffs is not None:
            repr_str = (
                "The true and estimated coefficients "
                + "(and their standard errors) are\n\n"
            )
            for i, coeff in enumerate(estimates[:n_alpha]):
                repr_str += f"   alpha({i + 1}): {true_coeffs[i]: > 10.3f}"
                repr_str += f"{coeff: > 10.3f}  ({stderrs[i]: > 10.3f})\n"
                repr_str += "\n"
            for i, coeff in enumerate(estimates[n_alpha:]):
                j = n_alpha + i
                repr_str += (
                    f"   base {i + 1}: {true_coeffs[j]: > 10.3f}  "
                    + f"{coeff: > 10.3f}  ({stderrs[j]: > 10.3f})\n"
                )
            print_stars(repr_str)
            discrepancy = npmaxabs(true_coeffs - estimates)
            print_stars(
                "The largest difference between true and estimated coefficients is"
                f" {discrepancy: .2e}"
            )
        else:
            repr_str = (
                "The estimated coefficients " + "(and their standard errors) are\n\n"
            )
            for i, coeff in enumerate(estimates[:n_alpha]):
                repr_str + f"{coeff: > 10.3f}  ({stderrs[i]: > 10.3f})\n"
                repr_str += "\n"
            for i, coeff in enumerate(estimates[n_alpha:]):
                j = n_alpha + i
                repr_str += f"{coeff: > 10.3f}  ({stderrs[j]: > 10.3f})\n"
            print_stars(repr_str)

        repr_str = "\nSpecification test:\n"
        repr_str += (
            "   the value of the test statistic is "
            + f"{self.test_statistic: > 10.3f}\n"
        )
        repr_str += (
            f"     for a chi2({self.ndf}), "
            + f"the p-value is {self.test_pvalue: > 10.3f}\n"
        )
        print_stars(repr_str)
        if true_coeffs is not None:
            return cast(float, discrepancy)
        return None



"""The components of the derivative of the entropy for the Choo and Siow homoskedastic model w/o singles.
"""

from typing import cast

import numpy as np
from bs_python_utils.bsnputils import ThreeArrays, TwoArrays, bs_error_abort

from cupid_matching.entropy import (
    EntropyFunctions,
    EntropyHessians,
    check_additional_parameters,
)
from cupid_matching.matching_utils import Matching


def _entropy_choo_siow_no_singles(
    muhat: Matching, deriv: int | None = 0
) -> (
    None
    | float
    | tuple[float, np.ndarray]
    | tuple[float, np.ndarray, np.ndarray, np.ndarray]
):
    """Returns the values of $\\mathcal{E}$ and the first (if `deriv` is 1 or 2) and second (if `deriv` is 2) derivatives
        for the Choo and Siow model w/o singles

    Args:
        muhat: a `Matching`
        deriv: if equal 1, we compute the first derivatives too;
               if equals 2, also the hessian

    Returns:
        the value of the generalized entropy
        if deriv = 1 or 2, the (X,Y) matrix of the first derivative of the entropy
        if deriv = 2, the (X,Y,X,Y) array of the second derivative
            wrt $(\\mu,\\mu)$
          and the (X,Y,X+Y) second derivatives
            wrt $(\\mu,(n,m))$
    """
    muxy, *_, n, m = muhat.unpack()

    logxy = np.log(muxy)
    logn = np.log(n)
    logm = np.log(m)

    val_entropy = -2.0 * np.sum(muxy * logxy) + np.sum(n * logn) + np.sum(m * logm)

    if deriv == 0:
        return cast(float, val_entropy)
    if deriv in [1, 2]:
        der_xy = -2.0 * (logxy + 1.0)
        if deriv == 1:
            return val_entropy, der_xy
        else:  # we compute_ the Hessians
            X, Y = muxy.shape
            der2_xyzt = np.zeros((X, Y, X, Y))
            der2_xyr = np.zeros((X, Y, X + Y))
            for x in range(X):
                for y in range(Y):
                    der2_xyzt[x, y, x, y] = -2.0 / muxy[x, y]
            return val_entropy, der_xy, der2_xyzt, der2_xyr
    else:
        bs_error_abort("deriv should be 0, 1, or 2")
        return None


def _der_entropy_choo_siow_no_singles_corrected(
    muhat: Matching, hessian: bool | None = False
) -> np.ndarray | ThreeArrays:
    """Returns the corrected first derivative of $\\mathcal{E}$ and the corrected second derivative (if `hessian` is True)
        for the Choo and Siow model w/o singles

    Args:
        muhat: a `Matching`
        hessian: if `True`, also compute_ the hessian

    Returns:
        the (X,Y) matrix of the first derivative of the entropy
        if hessian is True, the (X,Y,X,Y) array of the second derivative wrt $(\\mu,\\mu)$
          and the (X,Y,X+Y) second derivatives wrt $(\\mu,(n,m))$
    """
    muxy, *_ = muhat.unpack()
    n_households = np.sum(muxy)

    muxy_corr = muxy + (1.0 - muxy / n_households) / 2.0
    logxy = np.log(muxy_corr)

    der_xy = -2.0 * (logxy + 1.0)
    if not hessian:
        return der_xy
    else:  # we compute_ the Hessians
        X, Y = muxy.shape
        f_corr = 1.0 - 1.0 / n_households / 2.0
        derlogxy = f_corr / muxy_corr
        der2_xyzt = np.zeros((X, Y, X, Y))
        der2_xyr = np.zeros((X, Y, X + Y))
        for x in range(X):
            for y in range(Y):
                der2_xyzt[x, y, x, y] = -2.0 * derlogxy[x, y]
        return der_xy, der2_xyzt, der2_xyr


def e0_fun_choo_siow_no_singles(
    muhat: Matching,
    additional_parameters: list | None = None,
) -> np.ndarray:
    """Returns the values of $e_0$ for the Choo and Siow model w/o singles.

    Args:
        muhat: a `Matching`

    Returns:
        the `(X,Y)` matrix of the first derivative of the entropy
    """
    check_additional_parameters(0, additional_parameters)
    _, der_entropy = cast(
        tuple[float, np.ndarray], _entropy_choo_siow_no_singles(muhat, deriv=1)
    )
    return der_entropy


def e0_fun_choo_siow_no_singles_corrected(
    muhat: Matching,
    additional_parameters: list | None = None,
) -> np.ndarray:
    """Returns the values of $e_0$ for the Choo and Siow model,
        using the finite-sample correction $\\log(p+(1-p)/(2N))$

    Args:
        muhat: a `Matching`

    Returns:
        the (X,Y) matrix of the first derivative of the entropy
    """
    check_additional_parameters(0, additional_parameters)
    e0_val_corrected = _der_entropy_choo_siow_no_singles_corrected(muhat, hessian=False)
    return e0_val_corrected


def hessian_mumu_choo_siow_no_singles(
    muhat: Matching,
    additional_parameters: list | None = None,
) -> ThreeArrays:
    """Returns the hessian of $e_0$ wrt $(\\mu,\\mu)$ for the Choo and Siow model w/o singles.

    Args:
        muhat: a `Matching`

    Returns:
        the three components of the hessian  of the entropy wrt $(\\mu,\\mu)$
    """
    check_additional_parameters(0, additional_parameters)
    _, _, hessmumu, _ = cast(
        tuple[float, np.ndarray, np.ndarray, np.ndarray],
        _entropy_choo_siow_no_singles(muhat, deriv=2),
    )
    muxy, *_ = muhat.unpack()
    X, Y = muxy.shape
    hess_x = np.zeros((X, Y, Y))
    hess_y = np.zeros((X, Y, X))
    hess_xy = np.zeros((X, Y))
    for x in range(X):
        d2x = hessmumu[x, :, x, :]
        for y in range(Y):
            hess_xy[x, y] = d2x[y, y]
    return hess_x, hess_y, hess_xy


def hessian_mumu_choo_siow_no_singles_corrected(
    muhat: Matching,
    additional_parameters: list | None = None,
) -> ThreeArrays:
    """Returns the derivatives of the hessian of $e_0$ wrt $(\\mu,\\mu)$
        for the Choo and Siow model w/o singles, with the small sample correction

    Args:
        muhat: a `Matching`

    Returns:
        the three components of the hessian of the entropy wrt $(\\mu,\\mu)$
    """
    check_additional_parameters(0, additional_parameters)
    _, hessmumu, _ = cast(
        tuple[np.ndarray, np.ndarray, np.ndarray],
        _der_entropy_choo_siow_no_singles_corrected(muhat, hessian=True),
    )
    X, Y = muhat.muxy.shape
    hess_x = np.zeros((X, Y, Y))
    hess_y = np.zeros((X, Y, X))
    hess_xy = np.zeros((X, Y))
    for x in range(X):
        d2x = hessmumu[x, :, x, :]
        for y in range(Y):
            hess_xy[x, y] = d2x[y, y]
    return hess_x, hess_y, hess_xy


def hessian_mur_choo_siow_no_singles(
    muhat: Matching,
    additional_parameters: list | None = None,
) -> TwoArrays:
    """Returns the hessian of $e_0$ wrt $(\\mu,r)$ for the Choo and Siow model w/o singles.

    Args:
        muhat: a Matching

    Returns:
        the two components of the hessian of the entropy wrt $(\\mu,r)$
    """
    check_additional_parameters(0, additional_parameters)
    X, Y = muhat.muxy.shape
    hess_nx = np.zeros((X, Y))
    hess_my = np.zeros((X, Y))
    return hess_nx, hess_my


def hessian_mur_choo_siow_no_singles_corrected(
    muhat: Matching,
    additional_parameters: list | None = None,
) -> TwoArrays:
    """Returns the hessian of $e_0$ wrt $(\\mu,r)$ for the Choo and Siow model w/o singles, with the small sample correction

    Args:
        muhat: a `Matching`

    Returns:
        the two components of the hessian  of the entropy wrt $(\\mu,r)$
    """
    check_additional_parameters(0, additional_parameters)
    X, Y = muhat.muxy.shape
    hess_nx = np.zeros((X, Y))
    hess_my = np.zeros((X, Y))
    return hess_nx, hess_my


e0_derivative_choo_siow_no_singles = (
    hessian_mumu_choo_siow_no_singles,
    hessian_mur_choo_siow_no_singles,
)
e0_derivative_choo_siow_no_singles_corrected = (
    hessian_mumu_choo_siow_no_singles_corrected,
    hessian_mur_choo_siow_no_singles_corrected,
)

entropy_choo_siow_no_singles = EntropyFunctions(
    e0_fun=e0_fun_choo_siow_no_singles,
    hessian="provided",
    e0_derivative=cast(EntropyHessians, e0_derivative_choo_siow_no_singles),
    description="Choo and Siow homoskedastic  w/o singles with analytic Hessian",
)

entropy_choo_siow_no_singles_corrected = EntropyFunctions(
    e0_fun=e0_fun_choo_siow_no_singles_corrected,
    hessian="provided",
    e0_derivative=cast(EntropyHessians, e0_derivative_choo_siow_no_singles_corrected),
    description=(
        "Choo and Siow homoskedastic w/o singles with analytic Hessian and"
        " finite-sample correction"
    ),
)

entropy_choo_siow_no_singles_numeric = EntropyFunctions(
    e0_fun=e0_fun_choo_siow_no_singles,
    description="Choo and Siow homoskedastic w/o singles with numerical Hessian",
)

entropy_choo_siow_no_singles_corrected_numeric = EntropyFunctions(
    e0_fun=e0_fun_choo_siow_no_singles_corrected,
    description=(
        "Choo and Siow homoskedastic w/o singles with numerical Hessian and"
        " finite-sample correction"
    ),
)



""" example using the Choo and Siow homoskedastic model without singles"""

from typing import cast

import numpy as np
from bs_python_utils.bsutils import print_stars

from cupid_matching.choo_siow_no_singles import (
    entropy_choo_siow_no_singles,
    entropy_choo_siow_no_singles_corrected,
    entropy_choo_siow_no_singles_corrected_numeric,
    entropy_choo_siow_no_singles_numeric,
)
from cupid_matching.example_choo_siow import mde_estimate
from cupid_matching.model_classes import ChooSiowPrimitives
from cupid_matching.poisson_glm import choo_siow_poisson_glm


def create_choosiow_no_singles_population(
    X: int, Y: int, K: int, std_betas: float
) -> tuple[ChooSiowPrimitives, np.ndarray, np.ndarray]:
    """
    we simulate a Choo and Siow population
    with equal numbers of men and women of each type
    and random bases functions and coefficients

        Args:
         X: number of types of men
         Y: number of types of women
         K: random basis functions
         std_betas: the coefficients are drawn from a centered normal
                     with this standard deviation

        Returns:
            a `ChooSiowPrimitivesNoSingles` instance, the basis functions, and the coefficients
    """
    betas_true = std_betas * np.random.randn(K)
    phi_bases = np.random.randn(X, Y, K)
    n = np.ones(X)
    m = np.full(Y, X / Y)  # we need as many women as men overall
    Phi = phi_bases @ betas_true
    choo_siow_instance = ChooSiowPrimitives(Phi, n, m)
    return choo_siow_instance, phi_bases, betas_true


def demo_choo_siow_no_singles(
    n_households: int, X: int, Y: int, K: int, std_betas: float = 1.0
) -> tuple[float, float, float, float, float]:
    """run four MDE estimators and the Poisson estimator
    on randomly generated data

    Args:
        n_households: number of households
        X: number of types of men
        Y: number of types of women
        K: number of basis functions
        std_betas: the standard errors of their coefficients

    Returns:
        the discrepancies of the five estimators
    """
    choo_siow_instance, phi_bases, betas_true = create_choosiow_no_singles_population(
        X, Y, K, std_betas
    )
    mus_sim = choo_siow_instance.simulate(n_households)

    # we estimate using four variants of the minimum distance estimator
    mde_discrepancy = mde_estimate(
        mus_sim,
        phi_bases,
        betas_true,
        entropy_choo_siow_no_singles,
        no_singles=True,
        title="RESULTS FOR MDE WITH ANALYTICAL GRADIENT",
    )
    mde_discrepancy_numeric = mde_estimate(
        mus_sim,
        phi_bases,
        betas_true,
        entropy_choo_siow_no_singles_numeric,
        no_singles=True,
        title="RESULTS FOR MDE WITH NUMERICAL GRADIENT",
    )
    mde_discrepancy_corrected = mde_estimate(
        mus_sim,
        phi_bases,
        betas_true,
        entropy_choo_siow_no_singles_corrected,
        no_singles=True,
        title="RESULTS FOR THE CORRECTED MDE WITH ANALYTICAL GRADIENT",
    )
    mde_discrepancy_corrected_numeric = mde_estimate(
        mus_sim,
        phi_bases,
        betas_true,
        entropy_choo_siow_no_singles_corrected_numeric,
        no_singles=True,
        title="RESULTS FOR THE CORRECTED MDE WITH NUMERICAL GRADIENT",
    )

    # we also estimate using Poisson GLM
    print_stars("    RESULTS FOR POISSON   ")
    poisson_results = choo_siow_poisson_glm(mus_sim, phi_bases, no_singles=True)
    _, mux0_sim, mu0y_sim, n_sim, m_sim = mus_sim.unpack()
    # we normalize u_1 = 0
    u_true = -np.log(mux0_sim / n_sim)
    v_true = -np.log(mu0y_sim / m_sim) + u_true[0]
    u_true -= u_true[0]
    poisson_discrepancy = poisson_results.print_results(
        betas_true,
        u_true,
        v_true,
    )
    print("\n   (we normalized the utility u_1 at 0)\n")
    return (
        mde_discrepancy,
        mde_discrepancy_numeric,
        mde_discrepancy_corrected,
        mde_discrepancy_corrected_numeric,
        cast(float, poisson_discrepancy),
    )


if __name__ == "__main__":
    n_households = 1_000_000
    X, Y = 10, 15
    K = 8
    std_betas = 0.5
    (
        mde_discrepancy,
        mde_discrepancy_numeric,
        mde_discrepancy_corrected,
        mde_discrepancy_corrected_numeric,
        poisson_discrepancy,
    ) = demo_choo_siow_no_singles(n_households, X, Y, K, std_betas=std_betas)

    print_stars(
        "Largest absolute differences between the true and estimated coefficients:"
    )
    print(f"MDE:                            {mde_discrepancy: .2e}")
    print(f"MDE numeric:                    {mde_discrepancy_numeric: .2e}")
    print(f"MDE corrected:                  {mde_discrepancy_corrected: .2e}")
    print(f"MDE corrected numeric:          {mde_discrepancy_corrected_numeric: .2e}")
    print(f"Poisson:                        {poisson_discrepancy: .2e}")



"""This module contains some utility programs used by the package."""

from typing import Final, TypeAlias

import numpy as np
from bs_python_utils.bsutils import bs_error_abort

Nest: TypeAlias = list[int]
NestsList: TypeAlias = list[Nest]

# for numerical derivatives
_EPS: Final = 1e-6
_TWO_EPS: Final = 2.0 * _EPS


def make_XY_K_mat(xyk_array: np.ndarray) -> np.ndarray:
    """Reshapes an (X,Y,K) array to an (XY,K) matrix.

    Args:
        xyk_array: an (X, Y, K) array of bases

    Returns:
        the same,  (XY, K)-reshaped
    """
    X, Y, K = xyk_array.shape
    XY = X * Y
    xy_k_mat = np.zeros((XY, K))
    for k in range(K):
        xy_k_mat[:, k] = xyk_array[:, :, k].ravel()
    return xy_k_mat


def reshape4_to2(array4: np.ndarray) -> np.ndarray:
    """Reshapes an array (X,Y,Z,T) to a matrix (XY,ZT).

    Args:
        array4: an (X, Y, Z, T) array

    Returns:
        the same,  (XY, ZT)-reshaped
    """
    if array4.ndim != 4:
        bs_error_abort(f"array4 should have 4 dimensions not {array4.ndim}")
    X, Y, Z, T = array4.shape
    XY, ZT = X * Y, Z * T
    array2 = np.zeros((XY, ZT))
    xy = 0
    for x in range(X):
        for y in range(Y):
            array2[xy, :] = array4[x, y, :, :].ravel()
            xy += 1
    return array2


def change_indices(nests: NestsList):
    """subtracts 1 from the indices within the nest structure

    Args:
        nests: the nest structure

    Returns:
        a similar list
    """
    return [[nest_i - 1 for nest_i in nest] for nest in nests]


def find_nest_of(nests: NestsList, y: int) -> int:
    """find the index of the nest that contains y, or return -1

    Args:
        nests: a nest structure
        y: the type we are looking for

    Returns:
        the nest of y, or -1 if not found
    """
    for i_n, nest in enumerate(nests):
        if y in nest:
            return i_n
    return -1  # if not found



""" Estimates semilinear separable models with a given entropy function.
The entropy function and the surplus matrix must both be linear in the parameters.
"""

from typing import cast

import numpy as np
import scipy.stats as sts
from bs_python_utils.bsutils import print_stars

from cupid_matching.entropy import (
    EntropyFunctions,
    EntropyHessianMuMu,
    EntropyHessianMuR,
    EntropyHessians,
    numeric_hessian,
)
from cupid_matching.matching_utils import Matching, MatchingFunction
from cupid_matching.min_distance_utils import (
    MDEResults,
    check_args_mde,
    check_indep_phi_no_singles,
    compute_estimates,
    get_initial_weighting_matrix,
    get_optimal_weighting_matrix,
    make_D2_matrix,
    make_hessian_mde,
)
from cupid_matching.utils import make_XY_K_mat


def estimate_semilinear_mde(
    muhat: Matching,
    phi_bases: np.ndarray,
    entropy: EntropyFunctions,
    no_singles: bool = False,
    additional_parameters: list | None = None,
    initial_weighting_matrix: np.ndarray | None = None,
    verbose: bool = False,
) -> MDEResults:
    """
    Estimates the parameters of the distributions and of the base functions.

    Args:
        muhat: the observed `Matching`
        phi_bases: an (X, Y, K) array of bases
        entropy: an `EntropyFunctions` object
        no_singles: if `True`, only couples are observed
        additional_parameters: additional parameters of the distribution of errors,
            if any
        initial_weighting_matrix: if specified, used as the weighting matrix
            for the first step when `entropy.param_dependent` is `True`
        verbose: prints stuff if `True`

    Returns:
        an `MDEResults` instance

    Example:
        ```py
        # We simulate a Choo and Siow homoskedastic marriage market
        #  and we estimate a gender-heteroskedastic model on the simulated data.
        X, Y, K = 10, 20, 2
        n_households = int(1e6)
        lambda_true = np.random.randn(K)
        phi_bases = np.random.randn(X, Y, K)
        n = np.ones(X)
        m = np.ones(Y)
        Phi = phi_bases @ lambda_true
        choo_siow_instance = ChooSiowPrimitives(Phi, n, m)
        mus_sim = choo_siow_instance.simulate(n_households)
        choo_siow_instance.describe()

        entropy_model =  entropy_choo_siow_gender_heteroskedastic_numeric
        n_alpha = 1
        true_alpha = np.ones(n_alpha)
        true_coeffs = np.concatenate((true_alpha, lambda_true))

        print_stars(entropy_model.description)

        mde_results = estimate_semilinear_mde(
            mus_sim, phi_bases, entropy_model)

        mde_results.print_results(true_coeffs=true_coeffs, n_alpha=1)
        ```

    """
    X, Y, K = check_args_mde(muhat, phi_bases)
    XY = X * Y
    parameterized_entropy = entropy.parameter_dependent
    S_mat = get_initial_weighting_matrix(
        parameterized_entropy, initial_weighting_matrix, XY
    )

    phi_mat = make_XY_K_mat(phi_bases)

    # if there are no singles, we need to premultiply by the randomized double differencing matrix $D_2$
    if no_singles:
        D2_mat, rank_D2 = make_D2_matrix(X, Y)
        phi_mat = D2_mat @ phi_mat
        check_indep_phi_no_singles(phi_mat, X, Y)

    e0_vals = entropy.e0_fun(muhat, additional_parameters)
    e0_hat = e0_vals.ravel()

    # if there are no singles, we need to premultiply by the randomized double differencing matrix $D_2$
    if no_singles:
        e0_hat = D2_mat @ e0_hat

    if not parameterized_entropy:  # we only have e0(mu,r)
        n_pars = K
        hessian = entropy.hessian
        if hessian == "provided":  # we have the analytical hessian
            e0_derivative = cast(EntropyHessians, entropy.e0_derivative)
            hessian_components_mumu = e0_derivative[0](muhat, additional_parameters)
            hessian_components_mur = e0_derivative[1](muhat, additional_parameters)
        else:  # we use a numerical hessian
            hessian_components_mumu, hessian_components_mur = numeric_hessian(
                entropy,
                muhat,
                additional_parameters=additional_parameters,
            )

        hessians_both = make_hessian_mde(
            hessian_components_mumu, hessian_components_mur
        )

        # if there are no singles, we need to premultiply by the randomized double differencing matrix $D_2$
        if no_singles:
            S_mat = get_optimal_weighting_matrix(
                muhat, hessians_both, no_singles, D2_mat
            )
        else:
            S_mat = get_optimal_weighting_matrix(muhat, hessians_both)

        estimated_coefficients, varcov_coefficients = compute_estimates(
            phi_mat, S_mat, e0_hat
        )
        stderrs_coefficients = np.sqrt(np.diag(varcov_coefficients))
        est_Phi = phi_mat @ estimated_coefficients
        residuals = est_Phi + e0_hat
    else:  # parameterized entropy: e0(mu,r) + e(mu,r) . alpha
        e_fun = cast(MatchingFunction, entropy.e_fun)
        e_vals = e_fun(muhat, additional_parameters)
        e_hat = make_XY_K_mat(e_vals)

        # if there are no singles, we need to premultiply by the randomized double differencing matrix $D_2$
        if no_singles:
            e0_hat = D2_mat @ e0_hat

        F_hat = np.column_stack((e_hat, phi_mat))
        n_pars = e_hat.shape[1] + K

        # first pass with an initial weighting matrix
        first_coeffs, _ = compute_estimates(F_hat, cast(np.ndarray, S_mat), e0_hat)
        first_alpha = first_coeffs[:-K]

        if verbose:
            print_stars("First-stage estimates:")
            print(first_coeffs)

        # compute the efficient weighting matrix
        hessian = entropy.hessian
        if hessian == "provided":  # we have the analytical hessian
            e0_derivative = cast(EntropyHessians, entropy.e0_derivative)
            e_derivative = cast(EntropyHessians, entropy.e_derivative)
            e0_derivative_mumu = cast(EntropyHessianMuMu, e0_derivative[0])
            e0_derivative_mur = cast(EntropyHessianMuR, e0_derivative[1])
            e_derivative_mumu = cast(EntropyHessianMuMu, e_derivative[0])
            e_derivative_mur = cast(EntropyHessianMuR, e_derivative[1])
            hessian_components_mumu_e0 = e0_derivative_mumu(
                muhat, additional_parameters
            )
            hessian_components_mur_e0 = e0_derivative_mur(muhat, additional_parameters)
            hessian_components_mumu_e = e_derivative_mumu(muhat, additional_parameters)
            hessian_components_mur_e = e_derivative_mur(muhat, additional_parameters)

            if verbose:
                print_stars("First-stage estimates:")
                print(first_coeffs)

            hessian_components_mumu = (
                hessian_components_mumu_e0[i]
                + hessian_components_mumu_e[i] @ first_alpha
                for i in range(3)
            )
            hessian_components_mur = (
                hessian_components_mur_e0[i] + hessian_components_mur_e[i] @ first_alpha
                for i in range(3)
            )
        else:  # we use a numeric hessian
            hessian_components_mumu, hessian_components_mur = numeric_hessian(
                entropy,
                muhat,
                alpha=first_alpha,
                additional_parameters=additional_parameters,
            )
        hessians_both = make_hessian_mde(
            hessian_components_mumu, hessian_components_mur
        )

        # if there are no singles, we need to premultiply by the randomized double differencing matrix $D_2$
        if no_singles:
            S_mat = get_optimal_weighting_matrix(
                muhat, hessians_both, no_singles, D2_mat
            )
        else:
            S_mat = get_optimal_weighting_matrix(muhat, hessians_both)

        # second pass with the efficient weighting matrix
        estimated_coefficients, varcov_coefficients = compute_estimates(
            F_hat, S_mat, e0_hat
        )
        est_alpha, est_beta = (
            estimated_coefficients[:-K],
            estimated_coefficients[-K:],
        )
        stderrs_coefficients = np.sqrt(np.diag(varcov_coefficients))
        est_Phi = phi_mat @ est_beta
        residuals = est_Phi + e0_hat + e_hat @ est_alpha

    value_obj = residuals.T @ S_mat @ residuals
    ndf = rank_D2 - n_pars if no_singles else XY - n_pars
    test_stat = value_obj
    muxyhat, *_, nhat, mhat = muhat.unpack()
    n_individuals = np.sum(nhat) + np.sum(mhat)
    n_households = n_individuals - np.sum(muxyhat)

    est_Phi = est_Phi.reshape((X - 1, Y - 1)) if no_singles else est_Phi.reshape((X, Y))

    results = MDEResults(
        X=X,
        Y=Y,
        K=K,
        number_households=n_households,
        estimated_coefficients=estimated_coefficients,
        varcov_coefficients=varcov_coefficients,
        stderrs_coefficients=stderrs_coefficients,
        estimated_Phi=est_Phi,
        test_statistic=test_stat,
        ndf=ndf,
        test_pvalue=sts.chi2.sf(test_stat, ndf),
        parameterized_entropy=parameterized_entropy,
    )
    return results



from math import pow
from typing import Any, cast

import altair as alt
import numpy as np
import pandas as pd
import streamlit as st
from bs_python_utils.bsnputils import check_matrix, check_vector
from bs_python_utils.bsutils import bs_error_abort

from cupid_matching.matching_utils import Matching


def _make_profile(lambda_val: float, n: int, ncat: int) -> np.ndarray:
    """
    Creates a profile of numbers of individuals by type
    that changes exponentially

    Args:
        lambda_val: the exponent of the exponential
        n: the total number of individuals on that side of the market
        ncat: the number of types on that side of the market

    Returns:
        n_types: the number of individuals of each type
    """
    n1 = n * (lambda_val - 1.0) / (pow(lambda_val, ncat) - 1.0)
    n_types = n1 * np.logspace(base=lambda_val, start=0, stop=ncat - 1, num=ncat)
    return n_types


def make_margins(n: int, ncat: int, scenario: str = "Constant") -> np.ndarray:
    """generates the numbers by type on one side of the market

    Args:
        n: the total number of individuals on that side of the market
        ncat: the number of types on that side of the market
        scenario: "Constant", "Increasing", or "Decreasing" numbers as a function of type

    Returns:
        the numbers by type on this side
    """
    n_constant = n / ncat
    if scenario == "Constant":
        n_types = np.full(ncat, n_constant)
        return n_types
    elif scenario == "Increasing":
        lambda_val = pow(2.0, 1.0 / (ncat - 1))
    elif scenario == "Decreasing":
        lambda_val = pow(2.0, -1.0 / (ncat - 1))
    else:
        bs_error_abort(f"Unknown scenario {scenario}")
    n_types = _make_profile(lambda_val, n, ncat)
    return n_types


def table_estimates(
    coeff_names: list[str],
    true_coeffs: np.ndarray,
    estimates: np.ndarray,
    stderrs: np.ndarray,
) -> pd.DataFrame:
    """returns a table of the estimates

    Args:
        coeff_names: the names of the coefficients
        true_coeffs: the true values of the coefficients
        estimates: the estimated values of the coefficients
        stderrs: the standard errors of the estimates

    Returns:
        the table
    """
    st.write("The coefficients are:")
    df_coeffs_estimates = pd.DataFrame(
        {
            "True": true_coeffs,
            "Estimated": estimates,
            "Standard errors": stderrs,
        },
        index=coeff_names,
    )
    return df_coeffs_estimates


def plot_heatmap(
    mat: np.ndarray, str_format: str, str_tit: str | None = None
) -> alt.Chart:
    """Plots a heatmap of the matrix

    Args:
        mat: the matrix to plot
        str_tit: a title, if any

    Returns:
        the heatmap
    """
    ncat_men, ncat_women = check_matrix(mat)
    mat_arr = np.empty((mat.size, 4), dtype=int)
    imat = np.round(mat)
    mat_min = np.min(imat)
    i = 0
    for ix in range(ncat_men):
        for iy in range(ncat_women):
            m = imat[ix, iy]
            s = m - mat_min + 1
            mat_arr[i, :] = np.array([ix, iy, m, s])
            i += 1

    mat_df = pd.DataFrame(mat_arr, columns=["Men", "Women", "Value", "Size"])
    mat_df = mat_df.astype(
        dtype={"Men": int, "Women": int, "Value": int, "Size": float}
    )
    base = alt.Chart(mat_df).encode(x="Men:O", y=alt.Y("Women:O", sort="descending"))
    mat_map = base.mark_circle(opacity=0.4).encode(
        size=alt.Size("Size:Q", legend=None, scale=alt.Scale(range=[1000, 10000])),
        # color=alt.Color("Value:Q"),
        # tooltip=alt.Tooltip('Value', format=".2f")
    )
    text = base.mark_text(baseline="middle", fontSize=16).encode(
        text=alt.Text("Value:Q", format=str_format),
    )
    if str_tit is None:
        both = (mat_map + text).properties(width=500, height=500)
    else:
        both = (mat_map + text).properties(title=str_tit, width=400, height=400)
    return both


def _gender_singles(xvals: np.ndarray, str_gender: str) -> alt.Chart:
    """creates a histogram of singles for this side of the market

    Args:
        xvals: the numbers of singles by type
        str_gender: the side of the market, "men" or "women"

    Returns:
        the histogram
    """
    ncat = check_vector(xvals)
    if str_gender not in ["men", "women"]:
        bs_error_abort(f"{str_gender} is not a valid side")
    str_cat = "x" if str_gender == "men" else "y"
    str_val = f"Single {str_gender}"
    color_bar = "pink" if str_gender == "women" else "lightblue"
    source = pd.DataFrame({str_cat: np.arange(1, ncat + 1, dtype=int), str_val: xvals})

    g_bars = (
        alt.Chart(source)
        .mark_bar(color=color_bar)
        .encode(y=str_cat + ":O", x=str_val + ":Q")
    )
    return g_bars.properties(width=300, height=300)


def _plot_bars(mux0: np.ndarray, mu0y: np.ndarray) -> alt.Chart:
    """concatenates the two gender singles histograms

    Args:
        mux0: the numbers of single men by type
        mu0y: the numbers of single women by type

    Returns:
        the concatenated histogram
    """
    men_bars = _gender_singles(mux0, "men")
    women_bars = _gender_singles(mu0y, "women")
    return (men_bars & women_bars).properties(title="Singles")


def plot_matching(mus: Matching) -> alt.Chart:
    """generates the complete plot of matching patterns

    Args:
        mus: the matching patterns

    Returns:
        the plot
    """
    muxy, mux0, mu0y, _, _ = mus.unpack()
    plotxy = plot_heatmap(muxy, "d", str_tit="Marriages")
    plotsingles = _plot_bars(mux0, mu0y)
    return plotxy | plotsingles


def _convert_dataframe_to_csv(df: pd.DataFrame) -> str:
    """converts a DataFrame to a string in csv format"""
    return cast(str, df.to_csv().encode("utf-8"))


def _convert_arr_to_csv(arr: np.ndarray) -> str:
    """converts aa array to a string in csv format"""
    return _convert_dataframe_to_csv(pd.DataFrame(arr))


def _download_numpy_as_csv(arr: np.ndarray, file_name: str) -> None:
    """button to download an array to a csv file

    Args:
        arr: a Numpy array
        file_name: the array will be downloaded in file_name.csv

    Returns:
        nothing
    """
    csv = _convert_arr_to_csv(arr)
    _ = st.download_button(
        label=f"Download the {file_name} as a CSV file",
        data=csv,
        file_name=f"{file_name}.csv",
        mime="text/csv",
    )


def _download_dataframe_as_csv(df: pd.DataFrame, file_name: str) -> None:
    """button to download a DataFrame to a csv file

    Args:
        df: a Pandas DataFrame
        file_name: the datafrme will be downloaded in file_name.csv

    Returns:
        nothing
    """
    csv = _convert_dataframe_to_csv(df)
    _ = st.download_button(
        label=f"Download the {file_name} as a CSV file",
        data=csv,
        file_name=f"{file_name}.csv",
        mime="text/csv",
    )


def download_parameters_results(
    file_name: str, do_estimates: bool, pars_res: Any
) -> None:
    """formats the summary of the simulation (parameters and results) and offers to download
    the summary as a text file.

    Args:
        file_name: the summary will be downloaded in file_name
        do_estimates: if True, the two estimators were run and we also give the estimation results
        pars_res: a tuple with what we need to format the summary

    Returns:
        nothing
    """
    if do_estimates:
        (
            n_households,
            ncat_men,
            ncat_women,
            proportion_men,
            scenario_men,
            scenario_women,
            true_coeffs,
            coeff_names,
            muxy_sim,
            mux0_sim,
            mu0y_sim,
            n_sim,
            m_sim,
            df_mde,
            mde_test_results,
            df_poisson,
            df_u_estimates,
            df_v_estimates,
        ) = pars_res
        n_sim = cast(np.ndarray, n_sim)
        m_sim = cast(np.ndarray, m_sim)
        muxy_sim = cast(np.ndarray, muxy_sim)
        mux0_sim = cast(np.ndarray, mux0_sim)
        mu0y_sim = cast(np.ndarray, mu0y_sim)
        results_str = (
            f"You chose a model with {ncat_men} types of men and {ncat_women} types of"
            " women,\n"
        )
        results_str += (
            f"   with the number of men {scenario_men} and a number of wwomen"
            f" {scenario_women},\n"
        )
        results_str += (
            f"    and {n_households} households with a proportion of"
            f" {100*proportion_men:.1f} percent of men.\n\n"
        )
        results_str += "The numbers of men in each category are:\n"
        for x in range(ncat_men):
            results_str += f"{x+1}: {n_sim[x]: d}\n"
        results_str += "\n  the numbers of women in each category are:\n"
        for y in range(ncat_women):
            results_str += f"{y+1}: {m_sim[y]: d}\n"
        results_str += "\nYou chose the following coefficients:\n"
        for coeff, value in zip(coeff_names, true_coeffs, strict=True):
            results_str += f"{coeff}:  {value: >10.3f}\n"
        results_str += "\n\n"
        results_str += "This gives the following numbers of marriages:\n"
        for x in range(ncat_men):
            for y in range(ncat_women):
                results_str += f"{muxy_sim[x, y]: d}   "
            results_str += "\n"
        results_str += "\n\n"
        results_str += "The numbers of single men are:\n"
        for x in range(ncat_men):
            results_str += f"{x+1}: {mux0_sim[x]: d}\n"
        results_str += "\n\n"
        results_str += "The numbers of single women are:\n"
        for y in range(ncat_women):
            results_str += f"{y+1}: {mu0y_sim[y]: d}\n"
        results_str += "\n\n Minimum distance estimation gives\n"
        results_str += df_mde.to_string()
        specif_test_ndf, specif_test_stat, specif_test_pval = mde_test_results
        results_str += (
            f"\n\nSpecification test statistic: chi2({specif_test_ndf}) ="
            f" {specif_test_stat}\n"
        )
        results_str += f"     with  p-value {specif_test_pval}\n\n"
        results_str += "\n\n Poisson-GLM estimation gives\n"
        results_str += df_poisson.to_string()
        results_str += "\n\n  the expected utilities of men are:\n"
        results_str += df_u_estimates.to_string()
        results_str += "\n\n  the expected utilities of women are:\n"
        results_str += df_v_estimates.to_string()
    else:
        ncat_men, ncat_women, n_sim, m_sim, Phi, muxy_sim, mux0_sim, mu0y_sim = pars_res
        n_sim = cast(np.ndarray, n_sim)
        m_sim = cast(np.ndarray, m_sim)
        muxy_sim = cast(np.ndarray, muxy_sim)
        mux0_sim = cast(np.ndarray, mux0_sim)
        mu0y_sim = cast(np.ndarray, mu0y_sim)
        results_str = (
            f"You chose a model with {ncat_men} types of men and {ncat_women} types of"
            " women;\n"
        )
        results_str += "\n  the numbers of men in each category are:\n"
        for x in range(ncat_men):
            results_str += f"{x+1}: {n_sim[x]: d}\n"
        results_str += "\n  the numbers of women in each category are:\n"
        for y in range(ncat_women):
            results_str += f"{y+1}: {m_sim[y]: d}\n"
        results_str += "\nYou chose the following joint surplus matrix:\n"
        for x in range(ncat_men):
            for y in range(ncat_women):
                results_str += f"{Phi[x, y]: 10.2f} "
            results_str += "\n"
        results_str += "\n\n"
        results_str += "This gives the following numbers of marriages:\n"
        for x in range(ncat_men):
            for y in range(ncat_women):
                results_str += f"{muxy_sim[x, y]: d}   "
            results_str += "\n"
        results_str += "\n\n"
        results_str += "The numbers of single men are:\n"
        for x in range(ncat_men):
            results_str += f"{x+1}: {mux0_sim[x]: d}\n"
        results_str += "\n\n"
        results_str += "The numbers of single women are:\n"
        for y in range(ncat_women):
            results_str += f"{y+1}: {mu0y_sim[y]: d}\n"

    _ = st.download_button(
        label=f"Download summary to {file_name}",
        data=results_str,
        file_name=f"{file_name}",
        mime="text/csv",
    )



from typing import Any, cast

import numpy as np
import pandas as pd
import streamlit as st
from bs_python_utils.bsnputils import nprepeat_col, nprepeat_row

from cupid_matching.choo_siow import entropy_choo_siow
from cupid_matching.cupid_streamlit_utils import (
    download_parameters_results,
    make_margins,
    plot_heatmap,
    plot_matching,
    table_estimates,
)
from cupid_matching.min_distance import estimate_semilinear_mde
from cupid_matching.model_classes import ChooSiowPrimitives
from cupid_matching.poisson_glm import choo_siow_poisson_glm

summary_file_name = "summary.txt"  # simulation will be domwloaded there on request

st.title("Separable matching with transfers")

st.markdown(
    """
> This solves for equilibrium in a [Choo and Siow 2006](https://www.jstor.org/stable/10.1086/498585?seq=1) matching model
with perfectly transferable utility.

> If there are enough cells, it fits a quadratic specification of the joint surplus.

> You can also download a summary of your simulation.

> The app relies on the IPFP algorithm in [Galichon and Salanié 2022](https://academic.oup.com/restud/article-abstract/89/5/2600/6478301) and on the estimation methods in Galichon and Salanié (2023),

> which use the [cupid_matching](https://pypi.org/project/cupid_matching/) package.
"""
)

expander_bar = st.expander("More information")
expander_bar.markdown(
    """
The app lets you choose
* the total numbers  of men and women in a marriage market;
* the number of types $X$ and $Y$ of each;
* the proportions of men and women in each type, $n_x$ and $m_y$;
* and a joint surplus function $\\mathbb{\\Phi}$.

If $(X-1) \\times (Y-1) > 6$, you will be asked to specify the six coefficients $c_k$ of a quadratic specification for $\\mathbb{\\Phi}$:
"""
)
expander_bar.latex(r"\Phi_{xy}=c_0+c_1 x + c_2 y + c_3 x^2 + c_4 x y + c_5 y^2")
expander_bar.markdown("      and the profile of $n_x$ and $m_y$.")
expander_bar.markdown(
    """
If $(X-1) \\times (Y-1) \\leq 6$, there are not enough cells to estimate a quadratic specification;
you will be asked to input values of your choice for the $n_x, m_y$, and $\\Phi_{xy}$.
"""
)


expander_bar.markdown(
    "In all cases, the app plots the resulting joint surplus matrix $\\mathbb{\\Phi}$;"
)

expander_bar.markdown(
    """
Then it solves for the large market equilibrium in a simulated Choo and Siow market.

If $(X-1)\\times (Y-1) > 6$, it also fits the quadratic specification on the simulated equilibrium patterns using the two estimators in Galichon-Salanié (2023):

a minimum distance estimator and a Poisson GLM estimator.
"""
)

list_nhh = [10_000, 100_000, 1_000_000]
st.sidebar.subheader("First, choose the total number of households")
n_households = cast(int, st.sidebar.radio("Number of households", list_nhh))

list_ncat = [2, 3, 5, 10]
st.sidebar.subheader("Now choose the numbers of types of each gender")
ncat_men = cast(int, st.sidebar.radio("Number of categories of men", list_ncat))
ncat_women = cast(int, st.sidebar.radio("Number of categories of women", list_ncat))

enough_cells = (ncat_men - 1) * (ncat_women - 1) > 6

xvals = np.arange(ncat_men) + 1
yvals = np.arange(ncat_women) + 1

if enough_cells:
    st.sidebar.markdown(
        """
    By default there are as many men as women.
    You can also change the proportion.
    """
    )
    proportion_men = st.sidebar.slider(
        "Proportion of men", min_value=0.05, max_value=0.95, value=0.5
    )

    st.sidebar.markdown(
        """
    By default each category within a gender has the same number of individuals.
    You can also have the number increase by a factor two across categories, or decrease.
    """
    )

    list_scenarii = ["Constant", "Increasing", "Decreasing"]
    scenario_men = cast(
        str, st.sidebar.radio("Profile across categories for men", list_scenarii)
    )
    scenario_women = cast(
        str, st.sidebar.radio("Profile across categories for women", list_scenarii)
    )

    nx = make_margins(proportion_men, ncat_men, scenario_men)
    my = make_margins(1.0 - proportion_men, ncat_women, scenario_women)

    bases = np.zeros((ncat_men, ncat_women, 6))
    bases[:, :, 0] = 1.0
    xvals_mat = nprepeat_col(xvals, ncat_women)
    yvals_mat = nprepeat_row(yvals, ncat_men)
    bases[:, :, 1] = xvals_mat
    bases[:, :, 2] = yvals_mat
    bases[:, :, 3] = xvals_mat * xvals_mat
    bases[:, :, 4] = np.outer(xvals, yvals)
    bases[:, :, 5] = yvals_mat * yvals_mat

    st.sidebar.write("Finally, choose the coefficients of the 6 basis functions:")
    st.sidebar.latex(r"\Phi_{xy}=c_0+c_1 x + c_2 y + c_3 x^2 + c_4 x y + c_5 y^2")
    min_c = np.array([-3.0] + [-2.0 / ncat_men] * 5)
    max_c = np.array([3.0] + [2.0 / ncat_women] * 5)
    true_coeffs = np.zeros(6)
    coeff_names = [f"c[{i}]" for i in range(6)]

    if "randoms" not in st.session_state:
        random_coeffs = np.round(min_c + (max_c - min_c) * np.random.rand(6), 2)
        st.session_state.randoms = random_coeffs

    random_coeffs = st.session_state["randoms"]
    for i in range(6):
        val_i = float(random_coeffs[i])
        true_coeffs[i] = st.sidebar.slider(
            coeff_names[i], min_value=min_c[i], max_value=max_c[i], value=val_i
        )

    Phi = bases @ true_coeffs
else:
    st.markdown(
        """
        There are not enough cells to estimate a quadratic specification of the joint surplus;
        you need to choose the values of $n_x, m_y$, and $\\Phi_{xy}$.
        """
    )
    nx = np.zeros(ncat_men)
    my = np.zeros(ncat_women)
    st.subheader("Choose the numbers of men and women in each category")
    for iman in range(ncat_men):
        nx[iman] = st.slider(
            f"Number of men in category {iman+1}",
            min_value=1,
            max_value=10,
            step=1,
            value=5,
        )
    for iwoman in range(ncat_women):
        my[iwoman] = st.slider(
            f"Number of women in category {iwoman+1}",
            min_value=1,
            max_value=10,
            step=1,
            value=5,
        )
    Phi = np.zeros((ncat_men, ncat_women))
    for iman in range(ncat_men):
        for iwoman in range(ncat_women):
            Phi[iman, iwoman] = st.slider(
                f"Joint surplus in cell ({iman+1}, {iwoman+1})",
                min_value=-10.0,
                max_value=10.0,
                value=0.0,
            )


# if st.sidebar.button("Download Marriages"):
#     _download_file("marriages.csv")

# if st.sidebar.button("Download Singles"):
#     _download_file("singles.csv")

# if st.button("Download Minimum Distance Estimates"):
#     _download_file("min_distance_estimates.csv")

# if st.button("Download Poisson-GLM Estimates"):
#     _download_file("poisson_glm_estimates.csv")


st.subheader("Here is your joint surplus by categories:")
st.altair_chart(plot_heatmap(Phi, ".2f"))

cs_market = ChooSiowPrimitives(Phi, nx, my)

st.subheader(
    f"Here are the stable matching patterns in a sample of {n_households} households:"
)

mus_sim = cs_market.simulate(n_households)
muxy_sim, mux0_sim, mu0y_sim, n_sim, m_sim = mus_sim.unpack()

st.altair_chart(plot_matching(mus_sim))

do_estimates = False

if enough_cells:
    st.subheader(
        "Estimating the parameters of the quadratic joint surplus $\\mathbb{\\Phi}$"
    )

    if st.button("Estimate"):
        do_estimates = True
        col1, col2 = st.columns(2)
        with col1:
            st.markdown(
                "#### Below: the minimum distance estimator in Galichon and Salanié"
                " (2023)."
            )
            st.write("It also gives us a specification test.")
            mde_results = estimate_semilinear_mde(mus_sim, bases, entropy_choo_siow)
            mde_estimates = mde_results.estimated_coefficients
            mde_stderrs = mde_results.stderrs_coefficients

            df_mde = table_estimates(
                coeff_names, true_coeffs, mde_estimates, mde_stderrs
            )
            st.table(df_mde)

            specif_test_stat = round(mde_results.test_statistic, 2)
            specif_test_pval = round(mde_results.test_pvalue, 2)
            st.write(
                f"Test statistic: chi2({mde_results.ndf}) = {specif_test_stat} has"
                f" p-value {specif_test_pval}"
            )
            mde_test_results = (mde_results.ndf, specif_test_stat, specif_test_pval)

        with col2:
            st.markdown(
                "#### Here is the Poisson GLM estimator in Galichon and Salanié (2023);"
            )
            st.markdown(
                """
                as it requires numerical optimization, it is a bit slower.
                """
            )
            st.write(
                "It also gives us the estimates of the expected utilities $u_x$ and"
                " $v_y$."
            )

            pglm_results = choo_siow_poisson_glm(mus_sim, bases)

            u = pglm_results.estimated_u
            v = pglm_results.estimated_v
            pglm_estimates = pglm_results.estimated_beta
            pglm_stderrs = pglm_results.stderrs_beta

            df_poisson = table_estimates(
                coeff_names, true_coeffs, pglm_estimates, pglm_stderrs
            )
            st.table(df_poisson)

            x_names = [str(x) for x in range(ncat_men)]
            y_names = [str(y) for y in range(ncat_women)]

            st.write("The expected utilities are:")
            df_u_estimates = pd.DataFrame(
                {"Estimated": u, "True": -np.log(mux0_sim / n_sim)},
                index=x_names,
            )
            st.table(df_u_estimates)
            df_v_estimates = pd.DataFrame(
                {"Estimated": v, "True": -np.log(mu0y_sim / m_sim)},
                index=y_names,
            )
            st.table(df_v_estimates)
        pars_res = cast(
            Any,
            (
                n_households,
                ncat_men,
                ncat_women,
                proportion_men,
                scenario_men,
                scenario_women,
                true_coeffs,
                coeff_names,
                muxy_sim,
                mux0_sim,
                mu0y_sim,
                n_sim,
                m_sim,
                df_mde,
                mde_test_results,
                df_poisson,
                df_u_estimates,
                df_v_estimates,
            ),
        )

        download_parameters_results(summary_file_name, True, pars_res)
    else:
        pars_res = (
            ncat_men,
            ncat_women,
            n_sim,
            m_sim,
            Phi,
            muxy_sim,
            mux0_sim,
            mu0y_sim,
        )
        download_parameters_results(summary_file_name, False, pars_res)
else:
    pars_res = cast(
        Any, (ncat_men, ncat_women, n_sim, m_sim, Phi, muxy_sim, mux0_sim, mu0y_sim)
    )
    download_parameters_results(summary_file_name, False, pars_res)
